# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Helper functions to manage on-chain services"""

import datetime
import time
from typing import Callable, Dict, List, Optional, Tuple

from aea.crypto.base import Crypto, LedgerApi
from requests.exceptions import ConnectionError as RequestsConnectionError

from autonomy.chain.base import ServiceState, registry_contracts
from autonomy.chain.config import ChainType, ContractConfigs
from autonomy.chain.constants import (
    EVENT_VERIFICATION_TIMEOUT,
    GNOSIS_SAFE_MULTISIG_CONTRACT,
    SERVICE_MANAGER_CONTRACT,
    SERVICE_REGISTRY_CONTRACT,
)
from autonomy.chain.exceptions import (
    InstanceRegistrationFailed,
    ServiceDeployFailed,
    ServiceRegistrationFailed,
    TerminateServiceFailed,
    UnbondServiceFailed,
)
from autonomy.chain.mint import transact


DEFAULT_DEPLOY_PAYLOAD = "0x0000000000000000000000000000000000000000f48f2b2d2a534e402487b3ee7c18c33aec0fe5e4000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000"

ServiceInfo = Tuple[int, str, bytes, int, int, int, int, List[int]]


def get_default_delployment_payload() -> str:
    """Return default deployment payload."""
    return DEFAULT_DEPLOY_PAYLOAD + int(time.time()).to_bytes(32, "big").hex()


def get_agent_instances(
    ledger_api: LedgerApi, chain_type: ChainType, token_id: int
) -> Dict:
    """
    Get the list of agent instances.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param chain_type: Chain type
    :param token_id: Token ID pointing to the on-chain service
    :returns: number of agent instances and the list of registered addressed
    """

    return registry_contracts.service_registry.get_agent_instances(
        ledger_api=ledger_api,
        contract_address=ContractConfigs.get(SERVICE_REGISTRY_CONTRACT.name).contracts[
            chain_type
        ],
        service_id=token_id,
    )


def get_service_info(
    ledger_api: LedgerApi, chain_type: ChainType, token_id: int
) -> ServiceInfo:
    """
    Returns service info.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param chain_type: Chain type
    :param token_id: Token ID pointing to the on-chain service
    :returns: security deposit, multisig address, IPFS hash for config,
            threshold, max number of agent instances, number of agent instances,
            service state, list of cannonical agents
    """

    return registry_contracts.service_registry.get_service_information(
        ledger_api=ledger_api,
        contract_address=ContractConfigs.get(SERVICE_REGISTRY_CONTRACT.name).contracts[
            chain_type
        ],
        token_id=token_id,
    )


def wait_for_success_event(
    success_check: Callable[[], bool],
    message: str = "Timeout error",
    timeout: Optional[float] = None,
    sleep: float = 1.0,
) -> None:
    """Wait for success event."""

    timeout = timeout or EVENT_VERIFICATION_TIMEOUT
    deadline = datetime.datetime.now() + datetime.timedelta(seconds=timeout)
    while datetime.datetime.now() < deadline:
        if success_check():
            return
        time.sleep(sleep)
    raise TimeoutError(message)


def wait_for_agent_instance_registration(
    ledger_api: LedgerApi,
    chain_type: ChainType,
    service_id: int,
    instances: List[str],
    timeout: Optional[float] = None,
) -> None:
    """Wait for agent instance registration."""

    timeout = timeout or EVENT_VERIFICATION_TIMEOUT
    deadline = datetime.datetime.now() + datetime.timedelta(seconds=timeout)
    instance_check = set(instances)

    while datetime.datetime.now() < deadline:
        successful_instances = (
            registry_contracts.service_registry.verify_agent_instance_registration(
                ledger_api=ledger_api,
                contract_address=ContractConfigs.get(
                    SERVICE_REGISTRY_CONTRACT.name
                ).contracts[chain_type],
                service_id=service_id,
                instance_check=instance_check,
            )
        )

        instance_check = instance_check.difference(successful_instances)
        if len(instance_check) == 0:
            return

    raise TimeoutError(
        f"Could not verify the instance registration for {instance_check} in given time"
    )


def activate_service(
    ledger_api: LedgerApi,
    crypto: Crypto,
    chain_type: ChainType,
    service_id: int,
    timeout: Optional[float] = None,
) -> None:
    """
    Activate service.

    Once you have minted the service on-chain, you'll have to activate the service
    before you can proceed further.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param crypto: `aea.crypto.Crypto` object which has a funded key
    :param chain_type: Chain type
    :param service_id: Service ID retrieved after minting a service
    :param timeout: Time to wait for activation event to emit
    """

    (
        cost_of_bond,
        *_,
        service_state,
        _,
    ) = get_service_info(
        ledger_api=ledger_api, chain_type=chain_type, token_id=service_id
    )

    if service_state == ServiceState.NON_EXISTENT.value:
        raise ServiceRegistrationFailed("Service does not exist")

    if service_state != ServiceState.PRE_REGISTRATION.value:
        raise ServiceRegistrationFailed("Service must be inactive")

    try:
        tx = registry_contracts.service_manager.get_activate_registration_transaction(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                SERVICE_MANAGER_CONTRACT.name
            ).contracts[chain_type],
            owner=crypto.address,
            service_id=service_id,
            security_deposit=cost_of_bond,
            raise_on_try=True,
        )
        transact(ledger_api=ledger_api, crypto=crypto, tx=tx)
    except RequestsConnectionError as e:
        raise ServiceRegistrationFailed(
            "Service activation failed; Error connecting to the RPC"
        ) from e
    except ValueError as e:
        raise ServiceRegistrationFailed(f"Service activation failed; {e}") from e

    try:

        def success_check() -> bool:
            """Success check"""
            return (
                registry_contracts.service_registry.verify_service_has_been_activated(
                    ledger_api=ledger_api,
                    contract_address=ContractConfigs.get(
                        SERVICE_REGISTRY_CONTRACT.name
                    ).contracts[chain_type],
                    service_id=service_id,
                )
            )

        wait_for_success_event(
            success_check=success_check,
            message="Could not verify the service activation in given time",
            timeout=timeout,
        )
    except TimeoutError as e:
        raise ServiceRegistrationFailed(e) from e


def register_instance(  # pylint: disable=too-many-arguments
    ledger_api: LedgerApi,
    crypto: Crypto,
    chain_type: ChainType,
    service_id: int,
    instances: List[str],
    agent_ids: List[int],
    timeout: Optional[float] = None,
) -> None:
    """
    Register instance.

    Once you have a service with an active registration, you can register agent
    which will be a part of the service deployment. Using this method you can
    register maximum N amounts per agents, N being the number of slots for an agent
    with agent id being `agent_id`.

    Make sure the instance address you provide is not already a part of any service
    and not as same as the service owner.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param crypto: `aea.crypto.Crypto` object which has a funded key
    :param chain_type: Chain type
    :param service_id: Service ID retrieved after minting a service
    :param instances: Address of the agent instance
    :param agent_ids: Agent ID of the agent that you want this instance to be a part
                    of when deployed
    :param timeout: Time to wait for register instance event to emit
    """

    if len(agent_ids) != len(instances):
        raise InstanceRegistrationFailed(
            "Number of agent instances and agent IDs needs to be same"
        )

    (
        cost_of_bond,
        *_,
        service_state,
        _,
    ) = get_service_info(
        ledger_api=ledger_api,
        chain_type=chain_type,
        token_id=service_id,
    )

    if service_state == ServiceState.NON_EXISTENT.value:
        raise InstanceRegistrationFailed("Service does not exist")

    if service_state != ServiceState.ACTIVE_REGISTRATION.value:
        raise InstanceRegistrationFailed("Service needs to be in active state")

    security_deposit = cost_of_bond * len(agent_ids)

    try:
        tx = registry_contracts.service_manager.get_register_instance_transaction(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                SERVICE_MANAGER_CONTRACT.name
            ).contracts[chain_type],
            owner=crypto.address,
            service_id=service_id,
            instances=instances,
            agent_ids=agent_ids,
            security_deposit=security_deposit,
            raise_on_try=True,
        )

        transact(ledger_api=ledger_api, crypto=crypto, tx=tx)
    except RequestsConnectionError as e:
        raise InstanceRegistrationFailed(
            "Instance registration failed; Error connecting to the RPC"
        ) from e
    except ValueError as e:  # pragma: nocover
        raise InstanceRegistrationFailed(f"Instance registration failed; {e}") from e

    try:
        wait_for_agent_instance_registration(
            ledger_api=ledger_api,
            chain_type=chain_type,
            service_id=service_id,
            instances=instances,
            timeout=timeout,
        )
    except TimeoutError as e:
        raise InstanceRegistrationFailed(e) from e


def deploy_service(
    ledger_api: LedgerApi,
    crypto: Crypto,
    chain_type: ChainType,
    service_id: int,
    deployment_payload: Optional[str] = None,
    timeout: Optional[float] = None,
) -> None:
    """
    Deploy service.

    Using this method you can deploy a service on-chain once you have activated
    the service and registered the required agent instances.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param crypto: `aea.crypto.Crypto` object which has a funded key
    :param chain_type: Chain type
    :param service_id: Service ID retrieved after minting a service
    :param deployment_payload: Deployment payload to include when making the
                            deployment transaction
    :param timeout: Time to wait for deploy event to emit
    """
    deployment_payload = deployment_payload or get_default_delployment_payload()
    (
        *_,
        service_state,
        _,
    ) = get_service_info(
        ledger_api=ledger_api,
        chain_type=chain_type,
        token_id=service_id,
    )

    if service_state == ServiceState.NON_EXISTENT.value:
        raise ServiceDeployFailed("Service does not exist")

    if service_state != ServiceState.FINISHED_REGISTRATION.value:
        raise ServiceDeployFailed("Service needs to be in finished registration state")

    try:
        tx = registry_contracts.service_manager.get_service_deploy_transaction(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                SERVICE_MANAGER_CONTRACT.name
            ).contracts[chain_type],
            owner=crypto.address,
            service_id=service_id,
            gnosis_safe_multisig=ContractConfigs.get(
                GNOSIS_SAFE_MULTISIG_CONTRACT.name
            ).contracts[chain_type],
            deployment_payload=deployment_payload,
            raise_on_try=True,
        )
        transact(ledger_api=ledger_api, crypto=crypto, tx=tx)
    except RequestsConnectionError as e:
        raise ServiceDeployFailed(
            "Service deployment failed; Cannot connect to the RPC"
        ) from e
    except ValueError as e:  # pragma: nocover
        raise ServiceDeployFailed(f"Service deployment failed; {e}") from e

    try:

        def success_check() -> bool:
            """Success check"""
            return registry_contracts.service_registry.verify_service_has_been_deployed(
                ledger_api=ledger_api,
                contract_address=ContractConfigs.get(
                    SERVICE_REGISTRY_CONTRACT.name
                ).contracts[chain_type],
                service_id=service_id,
            )

        wait_for_success_event(
            success_check=success_check,
            message=f"Could not verify the service deployment for service {service_id} in given time",
            timeout=timeout,
        )
    except TimeoutError as e:
        raise ServiceDeployFailed(e) from e


def terminate_service(
    ledger_api: LedgerApi,
    crypto: Crypto,
    chain_type: ChainType,
    service_id: int,
) -> None:
    """
    Terminate service.

    Using this method you can terminate a service on-chain once you have activated
    the service and registered the required agent instances.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param crypto: `aea.crypto.Crypto` object which has a funded key
    :param chain_type: Chain type
    :param service_id: Service ID retrieved after minting a service
    """

    (
        *_,
        service_state,
        _,
    ) = get_service_info(
        ledger_api=ledger_api,
        chain_type=chain_type,
        token_id=service_id,
    )

    if service_state == ServiceState.NON_EXISTENT.value:
        raise TerminateServiceFailed("Service does not exist")

    if service_state == ServiceState.PRE_REGISTRATION.value:
        raise TerminateServiceFailed("Service not active")

    if service_state == ServiceState.TERMINATED_BONDED.value:
        raise TerminateServiceFailed("Service already terminated")

    try:
        tx = registry_contracts.service_manager.get_terminate_service_transaction(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                SERVICE_MANAGER_CONTRACT.name
            ).contracts[chain_type],
            owner=crypto.address,
            service_id=service_id,
            raise_on_try=True,
        )
        transact(ledger_api=ledger_api, crypto=crypto, tx=tx)
    except RequestsConnectionError as e:  # pragma: nocover
        raise TerminateServiceFailed(
            "Service termination failed; Cannot connect to the RPC"
        ) from e
    except ValueError as e:
        raise TerminateServiceFailed(f"Service termination failed; {e}") from e


def unbond_service(
    ledger_api: LedgerApi,
    crypto: Crypto,
    chain_type: ChainType,
    service_id: int,
) -> None:
    """
    Unbond service.

    Using this method you can unbond a service on-chain once you have terminated
    the service.

    :param ledger_api: `aea.crypto.LedgerApi` object for interacting with the chain
    :param crypto: `aea.crypto.Crypto` object which has a funded key
    :param chain_type: Chain type
    :param service_id: Service ID retrieved after minting a service
    """

    (
        *_,
        service_state,
        _,
    ) = get_service_info(
        ledger_api=ledger_api,
        chain_type=chain_type,
        token_id=service_id,
    )

    if service_state == ServiceState.NON_EXISTENT.value:
        raise UnbondServiceFailed("Service does not exist")

    if service_state != ServiceState.TERMINATED_BONDED.value:
        raise UnbondServiceFailed("Service needs to be in terminated-bonded state")

    try:
        tx = registry_contracts.service_manager.get_unbond_service_transaction(
            ledger_api=ledger_api,
            contract_address=ContractConfigs.get(
                SERVICE_MANAGER_CONTRACT.name
            ).contracts[chain_type],
            owner=crypto.address,
            service_id=service_id,
            raise_on_try=True,
        )
        transact(ledger_api=ledger_api, crypto=crypto, tx=tx)
    except RequestsConnectionError as e:  # pragma: nocover
        raise UnbondServiceFailed(
            "Service unbond failed; Cannot connect to the RPC"
        ) from e
    except ValueError as e:
        raise UnbondServiceFailed(f"Service unbond failed; {e}") from e



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test tools for CLI commands that use `autonomy.chain`"""

import os
import re
from pathlib import Path
from typing import Dict, List, Optional

import pytest
from aea.configurations.data_types import PackageId, PackageType, PublicId
from aea.crypto.base import Crypto, LedgerApi
from aea.helpers.base import IPFSHash
from aea_test_autonomy.configurations import ETHEREUM_KEY_DEPLOYER
from aea_test_autonomy.docker.base import skip_docker_tests
from aea_test_autonomy.fixture_helpers import registries_scope_class  # noqa: F401

from autonomy.chain.base import registry_contracts
from autonomy.chain.config import ChainType
from autonomy.chain.constants import (
    AGENT_REGISTRY_ADDRESS_LOCAL,
    COMPONENT_REGISTRY_ADDRESS_LOCAL,
    SERVICE_REGISTRY_ADDRESS_LOCAL,
)
from autonomy.chain.metadata import publish_metadata
from autonomy.chain.mint import (
    DEFAULT_NFT_IMAGE_HASH,
    UnitType,
    mint_component,
    mint_service,
)
from autonomy.chain.utils import parse_public_id_from_metadata, resolve_component_id
from autonomy.cli.helpers.chain import get_ledger_and_crypto_objects
from autonomy.cli.packages import get_package_manager

from tests.conftest import DATA_DIR
from tests.test_autonomy.test_cli.base import BaseCliTest


PACKAGES_DIR = DATA_DIR / "dummy_packages"
DUMMY_PACKAGE_MANAGER = get_package_manager(packages_dir=PACKAGES_DIR)
DUMMY_AUTHOR = "dummy_author"
DUMMY_PROTOCOL = PackageId(
    package_type=PackageType.PROTOCOL,
    public_id=PublicId(author=DUMMY_AUTHOR, name="dummy_protocol", version="0.1.0"),
)
DUMMY_CONTRACT = PackageId(
    package_type=PackageType.CONTRACT,
    public_id=PublicId(author=DUMMY_AUTHOR, name="dummy_contract", version="0.1.0"),
)
DUMMY_CONNECTION = PackageId(
    package_type=PackageType.CONNECTION,
    public_id=PublicId(author=DUMMY_AUTHOR, name="dummy_connection", version="0.1.0"),
)
DUMMY_SKILL = PackageId(
    package_type=PackageType.SKILL,
    public_id=PublicId(author=DUMMY_AUTHOR, name="dummy_skill", version="0.1.0"),
)
DUMMY_AGENT = PackageId(
    package_type=PackageType.AGENT,
    public_id=PublicId(author=DUMMY_AUTHOR, name="dummy_agent", version="0.1.0"),
)
DUMMY_SERVICE = PackageId(
    package_type=PackageType.SERVICE,
    public_id=PublicId(author=DUMMY_AUTHOR, name="dummy_service", version="0.1.0"),
)

DUMMY_HASH = "bafybei0000000000000000000000000000000000000000000000000000"
NUMBER_OF_SLOTS_PER_AGENT = 4
COST_OF_BOND_FOR_AGENT = 1000
THRESHOLD = 3
AGENT_ID = 1

DEFAULT_SERVICE_MINT_PARAMETERS = (
    "-a",
    str(AGENT_ID),
    "-n",
    str(NUMBER_OF_SLOTS_PER_AGENT),
    "-c",
    str(COST_OF_BOND_FOR_AGENT),
    "--threshold",
    str(THRESHOLD),
)


@skip_docker_tests
@pytest.mark.usefixtures("registries_scope_class")
class BaseChainInteractionTest(BaseCliTest):
    """Base chain interaction test"""

    ledger_api: LedgerApi
    crypto: Crypto
    chain_type: ChainType = ChainType.LOCAL

    key_file: Path = ETHEREUM_KEY_DEPLOYER

    def setup(self) -> None:
        """Setup test."""
        super().setup()
        self.cli_runner.mix_stderr = False

    @classmethod
    def setup_class(cls) -> None:
        """Setup class."""
        super().setup_class()

        cls.ledger_api, cls.crypto = get_ledger_and_crypto_objects(
            chain_type=cls.chain_type,
            key=cls.key_file,
        )

    @staticmethod
    def extract_token_id_from_output(output: str) -> int:
        """Extract token ID from output"""

        matches = re.findall(r"Token ID: (\d+)", output)
        assert len(matches) > 0

        *_, token_id = matches
        return int(token_id)

    def verify_owner_address(
        self, token_id: int, owner: str, package_id: PackageId
    ) -> None:
        """Verify minted token id"""

        if package_id.package_type == PackageType.SERVICE:
            on_chain_owner = (
                registry_contracts.service_registry.get_instance(
                    ledger_api=self.ledger_api,
                    contract_address=SERVICE_REGISTRY_ADDRESS_LOCAL,
                )
                .functions.ownerOf(token_id)
                .call()
            )
        elif package_id.package_type == PackageType.AGENT:
            on_chain_owner = (
                registry_contracts.component_registry.get_instance(
                    ledger_api=self.ledger_api,
                    contract_address=AGENT_REGISTRY_ADDRESS_LOCAL,
                )
                .functions.ownerOf(token_id)
                .call()
            )
        else:
            on_chain_owner = (
                registry_contracts.component_registry.get_instance(
                    ledger_api=self.ledger_api,
                    contract_address=COMPONENT_REGISTRY_ADDRESS_LOCAL,
                )
                .functions.ownerOf(token_id)
                .call()
            )

        assert (
            owner.lower() == on_chain_owner.lower()
        ), f"Owner verification failed; Owner address={owner}; On chain owner address={on_chain_owner}"

    def verify_minted_token_id(self, token_id: int, package_id: PackageId) -> None:
        """Verify minted token id"""

        is_agent = package_id.package_type == PackageType.AGENT
        is_service = package_id.package_type == PackageType.SERVICE

        if is_service:
            contract_address = SERVICE_REGISTRY_ADDRESS_LOCAL
        elif is_agent:
            contract_address = AGENT_REGISTRY_ADDRESS_LOCAL
        else:
            contract_address = COMPONENT_REGISTRY_ADDRESS_LOCAL

        metadata = resolve_component_id(
            ledger_api=self.ledger_api,
            contract_address=contract_address,
            token_id=token_id,
            is_agent=is_agent,
            is_service=is_service,
        )

        minted_public_id = parse_public_id_from_metadata(metadata["name"])
        assert package_id.public_id.to_any() == minted_public_id.to_any(), (
            minted_public_id,
            package_id.public_id,
        )

    @staticmethod
    def verify_and_remove_metadata_file(token_id: int) -> None:
        """Verify and remove the metadata file."""

        metadata_file = Path(f"{token_id}.json").resolve()
        assert metadata_file.exists()
        os.remove(metadata_file)

    def mint_component(
        self,
        package_id: PackageId,
        dependencies: Optional[List[int]] = None,
        service_mint_parameters: Optional[Dict] = None,
    ) -> int:
        """Mint component and return token id"""

        package_path = DUMMY_PACKAGE_MANAGER.package_path_from_package_id(
            package_id=package_id
        )
        metadata_hash, _ = publish_metadata(
            package_id=package_id,
            package_path=package_path,
            nft=IPFSHash(DEFAULT_NFT_IMAGE_HASH),
            description="Dummy package for testing",
        )
        if package_id.package_type == PackageType.SERVICE:
            assert (
                service_mint_parameters is not None
            ), "Please provide service mint parameters"
            token_id = mint_service(
                ledger_api=self.ledger_api,
                crypto=self.crypto,
                metadata_hash=metadata_hash,
                chain_type=ChainType.LOCAL,
                **service_mint_parameters,
            )
        else:
            token_id = mint_component(
                ledger_api=self.ledger_api,
                crypto=self.crypto,
                metadata_hash=metadata_hash,
                chain_type=ChainType.LOCAL,
                component_type=(
                    UnitType.AGENT
                    if package_id.package_type == PackageType.AGENT
                    else UnitType.COMPONENT
                ),
                dependencies=dependencies,
            )

        assert isinstance(token_id, int)
        return token_id



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 valory
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Serialization module for ipfs protocol."""

# pylint: disable=too-many-statements,too-many-locals,no-member,too-few-public-methods,redefined-builtin
from typing import Any, Dict, cast

from aea.mail.base_pb2 import DialogueMessage
from aea.mail.base_pb2 import Message as ProtobufMessage
from aea.protocols.base import Message, Serializer

from packages.valory.protocols.ipfs import ipfs_pb2
from packages.valory.protocols.ipfs.message import IpfsMessage


class IpfsSerializer(Serializer):
    """Serialization for the 'ipfs' protocol."""

    @staticmethod
    def encode(msg: Message) -> bytes:
        """
        Encode a 'Ipfs' message into bytes.

        :param msg: the message object.
        :return: the bytes.
        """
        msg = cast(IpfsMessage, msg)
        message_pb = ProtobufMessage()
        dialogue_message_pb = DialogueMessage()
        ipfs_msg = ipfs_pb2.IpfsMessage()

        dialogue_message_pb.message_id = msg.message_id
        dialogue_reference = msg.dialogue_reference
        dialogue_message_pb.dialogue_starter_reference = dialogue_reference[0]
        dialogue_message_pb.dialogue_responder_reference = dialogue_reference[1]
        dialogue_message_pb.target = msg.target

        performative_id = msg.performative
        if performative_id == IpfsMessage.Performative.STORE_FILES:
            performative = ipfs_pb2.IpfsMessage.Store_Files_Performative()  # type: ignore
            files = msg.files
            performative.files.update(files)
            if msg.is_set("timeout"):
                performative.timeout_is_set = True
                timeout = msg.timeout
                performative.timeout = timeout
            ipfs_msg.store_files.CopyFrom(performative)
        elif performative_id == IpfsMessage.Performative.IPFS_HASH:
            performative = ipfs_pb2.IpfsMessage.Ipfs_Hash_Performative()  # type: ignore
            ipfs_hash = msg.ipfs_hash
            performative.ipfs_hash = ipfs_hash
            ipfs_msg.ipfs_hash.CopyFrom(performative)
        elif performative_id == IpfsMessage.Performative.GET_FILES:
            performative = ipfs_pb2.IpfsMessage.Get_Files_Performative()  # type: ignore
            ipfs_hash = msg.ipfs_hash
            performative.ipfs_hash = ipfs_hash
            if msg.is_set("timeout"):
                performative.timeout_is_set = True
                timeout = msg.timeout
                performative.timeout = timeout
            ipfs_msg.get_files.CopyFrom(performative)
        elif performative_id == IpfsMessage.Performative.FILES:
            performative = ipfs_pb2.IpfsMessage.Files_Performative()  # type: ignore
            files = msg.files
            performative.files.update(files)
            ipfs_msg.files.CopyFrom(performative)
        elif performative_id == IpfsMessage.Performative.ERROR:
            performative = ipfs_pb2.IpfsMessage.Error_Performative()  # type: ignore
            reason = msg.reason
            performative.reason = reason
            ipfs_msg.error.CopyFrom(performative)
        else:
            raise ValueError("Performative not valid: {}".format(performative_id))

        dialogue_message_pb.content = ipfs_msg.SerializeToString()

        message_pb.dialogue_message.CopyFrom(dialogue_message_pb)
        message_bytes = message_pb.SerializeToString()
        return message_bytes

    @staticmethod
    def decode(obj: bytes) -> Message:
        """
        Decode bytes into a 'Ipfs' message.

        :param obj: the bytes object.
        :return: the 'Ipfs' message.
        """
        message_pb = ProtobufMessage()
        ipfs_pb = ipfs_pb2.IpfsMessage()
        message_pb.ParseFromString(obj)
        message_id = message_pb.dialogue_message.message_id
        dialogue_reference = (
            message_pb.dialogue_message.dialogue_starter_reference,
            message_pb.dialogue_message.dialogue_responder_reference,
        )
        target = message_pb.dialogue_message.target

        ipfs_pb.ParseFromString(message_pb.dialogue_message.content)
        performative = ipfs_pb.WhichOneof("performative")
        performative_id = IpfsMessage.Performative(str(performative))
        performative_content = dict()  # type: Dict[str, Any]
        if performative_id == IpfsMessage.Performative.STORE_FILES:
            files = ipfs_pb.store_files.files
            files_dict = dict(files)
            performative_content["files"] = files_dict
            if ipfs_pb.store_files.timeout_is_set:
                timeout = ipfs_pb.store_files.timeout
                performative_content["timeout"] = timeout
        elif performative_id == IpfsMessage.Performative.IPFS_HASH:
            ipfs_hash = ipfs_pb.ipfs_hash.ipfs_hash
            performative_content["ipfs_hash"] = ipfs_hash
        elif performative_id == IpfsMessage.Performative.GET_FILES:
            ipfs_hash = ipfs_pb.get_files.ipfs_hash
            performative_content["ipfs_hash"] = ipfs_hash
            if ipfs_pb.get_files.timeout_is_set:
                timeout = ipfs_pb.get_files.timeout
                performative_content["timeout"] = timeout
        elif performative_id == IpfsMessage.Performative.FILES:
            files = ipfs_pb.files.files
            files_dict = dict(files)
            performative_content["files"] = files_dict
        elif performative_id == IpfsMessage.Performative.ERROR:
            reason = ipfs_pb.error.reason
            performative_content["reason"] = reason
        else:
            raise ValueError("Performative not valid: {}.".format(performative_id))

        return IpfsMessage(
            message_id=message_id,
            dialogue_reference=dialogue_reference,
            target=target,
            performative=performative,
            **performative_content
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test metadata helpers."""

import tempfile
from pathlib import Path
from unittest import mock

import pytest
from aea.configurations.constants import DEFAULT_README_FILE
from aea.configurations.data_types import PackageId, PublicId
from aea.helpers.base import IPFSHash

from autonomy.chain.metadata import NFTHashOrPath, publish_metadata, serialize_metadata
from autonomy.chain.mint import DEFAULT_NFT_IMAGE_HASH

from tests.conftest import DATA_DIR
from tests.test_autonomy.test_chain.base import DUMMY_HASH


def test_serialize_metadata() -> None:
    """Test serialize metadata."""
    expected_string = """{"name":"skill/author/name/latest","description":"Some package","code_uri":"ipfs://bafybei0000000000000000000000000000000000000000000000000000","image":"ipfs://bafybeiggnad44tftcrenycru2qtyqnripfzitv5yume4szbkl33vfd4abm","attributes":[{"trait_type":"version","value":"latest"}]}"""
    metadata_string = serialize_metadata(
        package_hash=DUMMY_HASH,
        package_id=PackageId(
            package_type="skill", public_id=PublicId(author="author", name="name")
        ),
        description="Some package",
        nft_image_hash=DEFAULT_NFT_IMAGE_HASH,
    )

    assert metadata_string == expected_string


@pytest.mark.parametrize(
    "nft",
    (
        IPFSHash(DEFAULT_NFT_IMAGE_HASH),
        DATA_DIR / "nft.png",
    ),
)
def test_publish_metadata(nft: NFTHashOrPath) -> None:
    """Test publish metadata tool with dummy config."""

    expected_hash = "0x338a96df1d172c1d6f11a0460d075d80b62eea67134b2d0a52a7df34495d45d1"
    with mock.patch(
        "autonomy.chain.metadata.IPFSHashOnly.get", return_value=DUMMY_HASH
    ):
        with tempfile.TemporaryDirectory() as temp_dir:
            package_path = Path(temp_dir)
            (package_path / DEFAULT_README_FILE).write_text("Description")

            metadata_hash, _ = publish_metadata(
                package_id=PackageId(
                    package_type="skill",
                    public_id=PublicId(author="author", name="name"),
                ),
                package_path=package_path,
                nft=nft,
                description="",
            )

    assert metadata_hash == expected_hash



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Templates for FSM components"""

from dataclasses import dataclass


@dataclass
class ROUNDS:
    """Template for rounds.py"""

    FILENAME = "rounds.py"

    HEADER = """\
    \"\"\"This package contains the rounds of {AbciApp}.\"\"\"

    from enum import Enum
    from typing import Dict, List, Optional, Set, Tuple

    from packages.valory.skills.abstract_round_abci.base import (
        AbciApp,
        AbciAppTransitionFunction,
        AbstractRound,
        AppState,
        BaseSynchronizedData,
        DegenerateRound,
        EventToTimeout,
    )

    from packages.{author}.skills.{skill_name}.payloads import (
        {payloads},
    )


    class Event(Enum):
        \"\"\"{AbciApp} Events\"\"\"

        {events}


    class SynchronizedData(BaseSynchronizedData):
        \"\"\"
        Class to represent the synchronized data.

        This data is replicated by the tendermint application.
        \"\"\"

    """

    ROUND_CLS = """\
    class {RoundCls}({ABCRoundCls}):
        \"\"\"{RoundCls}\"\"\"

        payload_class = {PayloadCls}
        payload_attribute = ""  # TODO: update
        synchronized_data_class = SynchronizedData

        {todo_abstract_round_cls}

        def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
            \"\"\"Process the end of the block.\"\"\"
            raise NotImplementedError

        def check_payload(self, payload: {PayloadCls}) -> None:
            \"\"\"Check payload.\"\"\"
            raise NotImplementedError

        def process_payload(self, payload: {PayloadCls}) -> None:
            \"\"\"Process payload.\"\"\"
            raise NotImplementedError

    """

    DEGENERATE_ROUND_CLS = """\
    class {RoundCls}({ABCRoundCls}):
        \"\"\"{RoundCls}\"\"\"

    """

    ABCI_APP_CLS = """\
    class {AbciApp}(AbciApp[Event]):
        \"\"\"{AbciApp}\"\"\"

        initial_round_cls: AppState = {initial_round_cls}
        initial_states: Set[AppState] = {initial_states}
        transition_function: AbciAppTransitionFunction = {transition_function}
        final_states: Set[AppState] = {final_states}
        event_to_timeout: EventToTimeout = {{}}
        cross_period_persisted_keys: Set[str] = []
        db_pre_conditions: Dict[AppState, Set[str]] = {{
            {db_pre_conditions}
        }}
        db_post_conditions: Dict[AppState, Set[str]] = {{
            {db_post_conditions}
        }}
    """


@dataclass
class BEHAVIOURS:
    """Template for behaviours.py"""

    FILENAME = "behaviours.py"

    HEADER = """\
    \"\"\"This package contains round behaviours of {AbciApp}.\"\"\"

    from abc import ABC
    from typing import Generator, Set, Type, cast

    from packages.valory.skills.abstract_round_abci.base import AbstractRound
    from packages.valory.skills.abstract_round_abci.behaviours import (
        AbstractRoundBehaviour,
        BaseBehaviour,
    )

    from packages.{author}.skills.{skill_name}.models import Params
    from packages.{author}.skills.{skill_name}.rounds import (
        SynchronizedData,
        {AbciApp},
        {rounds},
    )
    from packages.{author}.skills.{skill_name}.rounds import (
        {payloads},
    )

    """

    BASE_BEHAVIOUR_CLS = """\
    class {BaseBehaviourCls}(BaseBehaviour, ABC):
        \"\"\"Base behaviour for the {skill_name} skill.\"\"\"

        @property
        def synchronized_data(self) -> SynchronizedData:
            \"\"\"Return the synchronized data.\"\"\"
            return cast(SynchronizedData, super().synchronized_data)

        @property
        def params(self) -> Params:
            \"\"\"Return the params.\"\"\"
            return cast(Params, super().params)

    """

    BEHAVIOUR_CLS = """\
    class {BehaviourCls}({BaseBehaviourCls}):
        \"\"\"{BehaviourCls}\"\"\"

        matching_round: Type[AbstractRound] = {matching_round}

        # TODO: implement logic required to set payload content for synchronization
        def async_act(self) -> Generator:
            \"\"\"Do the act, supporting asynchronous execution.\"\"\"

            with self.context.benchmark_tool.measure(self.behaviour_id).local():
                sender = self.context.agent_address
                payload = {PayloadCls}(sender=sender, content=...)

            with self.context.benchmark_tool.measure(self.behaviour_id).consensus():
                yield from self.send_a2a_transaction(payload)
                yield from self.wait_until_round_end()

            self.set_done()

    """

    ROUND_BEHAVIOUR_CLS = """\
    class {RoundBehaviourCls}(AbstractRoundBehaviour):
        \"\"\"{RoundBehaviourCls}\"\"\"

        initial_behaviour_cls = {InitialBehaviourCls}
        abci_app_cls = {AbciApp}  # type: ignore
        behaviours: Set[Type[BaseBehaviour]] = {round_behaviours}
    """


@dataclass
class PAYLOADS:
    """Template for payloads.py"""

    FILENAME = "payloads.py"

    HEADER = """\
    \"\"\"This module contains the transaction payloads of the {AbciApp}.\"\"\"

    from dataclasses import dataclass

    from packages.valory.skills.abstract_round_abci.base import BaseTxPayload

    """

    PAYLOAD_CLS = """\
    @dataclass(frozen=True)
    class {PayloadCls}(BaseTxPayload):
        \"\"\"Represent a transaction payload for the {RoundCls}.\"\"\"

        # TODO: define your attributes

    """


@dataclass
class MODELS:
    """Template for models.py"""

    FILENAME = "models.py"

    HEADER = """\
    \"\"\"This module contains the shared state for the abci skill of {AbciApp}.\"\"\"

    from packages.valory.skills.abstract_round_abci.models import BaseParams
    from packages.valory.skills.abstract_round_abci.models import (
        BenchmarkTool as BaseBenchmarkTool,
    )
    from packages.valory.skills.abstract_round_abci.models import Requests as BaseRequests
    from packages.valory.skills.abstract_round_abci.models import (
        SharedState as BaseSharedState,
    )
    from packages.{author}.skills.{skill_name}.rounds import {AbciApp}


    class SharedState(BaseSharedState):
        \"\"\"Keep the current shared state of the skill.\"\"\"

        abci_app_cls = {AbciApp}


    Params = BaseParams
    Requests = BaseRequests
    BenchmarkTool = BaseBenchmarkTool
    """


@dataclass
class HANDLERS:
    """Template for handlers.py"""

    FILENAME = "handlers.py"

    HEADER = """\
    \"\"\"This module contains the handlers for the skill of {AbciApp}.\"\"\"

    from packages.valory.skills.abstract_round_abci.handlers import (
        ABCIRoundHandler as BaseABCIRoundHandler,
    )
    from packages.valory.skills.abstract_round_abci.handlers import (
        ContractApiHandler as BaseContractApiHandler,
    )
    from packages.valory.skills.abstract_round_abci.handlers import (
        HttpHandler as BaseHttpHandler,
    )
    from packages.valory.skills.abstract_round_abci.handlers import (
        IpfsHandler as BaseIpfsHandler,
    )
    from packages.valory.skills.abstract_round_abci.handlers import (
        LedgerApiHandler as BaseLedgerApiHandler,
    )
    from packages.valory.skills.abstract_round_abci.handlers import (
        SigningHandler as BaseSigningHandler,
    )
    from packages.valory.skills.abstract_round_abci.handlers import (
        TendermintHandler as BaseTendermintHandler,
    )


    ABCIHandler = BaseABCIRoundHandler
    HttpHandler = BaseHttpHandler
    SigningHandler = BaseSigningHandler
    LedgerApiHandler = BaseLedgerApiHandler
    ContractApiHandler = BaseContractApiHandler
    TendermintHandler = BaseTendermintHandler
    IpfsHandler = BaseIpfsHandler
    """


@dataclass
class DIALOGUES:
    """Template for dialogues.py"""

    FILENAME = "dialogues.py"

    HEADER = """\
    \"\"\"This module contains the dialogues of the {AbciApp}.\"\"\"

    from packages.valory.skills.abstract_round_abci.dialogues import (
        AbciDialogue as BaseAbciDialogue,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        AbciDialogues as BaseAbciDialogues,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        ContractApiDialogue as BaseContractApiDialogue,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        ContractApiDialogues as BaseContractApiDialogues,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        HttpDialogue as BaseHttpDialogue,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        HttpDialogues as BaseHttpDialogues,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        IpfsDialogue as BaseIpfsDialogue,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        IpfsDialogues as BaseIpfsDialogues,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        LedgerApiDialogue as BaseLedgerApiDialogue,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        LedgerApiDialogues as BaseLedgerApiDialogues,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        SigningDialogue as BaseSigningDialogue,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        SigningDialogues as BaseSigningDialogues,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        TendermintDialogue as BaseTendermintDialogue,
    )
    from packages.valory.skills.abstract_round_abci.dialogues import (
        TendermintDialogues as BaseTendermintDialogues,
    )


    AbciDialogue = BaseAbciDialogue
    AbciDialogues = BaseAbciDialogues


    HttpDialogue = BaseHttpDialogue
    HttpDialogues = BaseHttpDialogues


    SigningDialogue = BaseSigningDialogue
    SigningDialogues = BaseSigningDialogues


    LedgerApiDialogue = BaseLedgerApiDialogue
    LedgerApiDialogues = BaseLedgerApiDialogues


    ContractApiDialogue = BaseContractApiDialogue
    ContractApiDialogues = BaseContractApiDialogues


    TendermintDialogue = BaseTendermintDialogue
    TendermintDialogues = BaseTendermintDialogues


    IpfsDialogue = BaseIpfsDialogue
    IpfsDialogues = BaseIpfsDialogues
    """



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests package for the 'valory/tendermint' protocol."""
import logging
from typing import Type
from unittest import mock

import pytest
from _pytest.logging import LogCaptureFixture  # type: ignore
from aea.common import Address
from aea.mail.base import Envelope
from aea.protocols.base import Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea.protocols.dialogue.base import DialogueLabel

from packages.valory.protocols.tendermint.custom_types import ErrorCode
from packages.valory.protocols.tendermint.dialogues import (
    TendermintDialogue,
    TendermintDialogues,
)
from packages.valory.protocols.tendermint.message import (
    TendermintMessage,
    _default_logger,
)


@pytest.mark.parametrize(
    "msg",
    [
        TendermintMessage(
            performative=TendermintMessage.Performative.GET_GENESIS_INFO,  # type: ignore
            query="",
        ),
        TendermintMessage(
            performative=TendermintMessage.Performative.GET_RECOVERY_PARAMS,  # type: ignore
            query="",
        ),
        TendermintMessage(
            performative=TendermintMessage.Performative.GENESIS_INFO,  # type: ignore
            info="",
        ),
        TendermintMessage(
            performative=TendermintMessage.Performative.RECOVERY_PARAMS,  # type: ignore
            params="",
        ),
        TendermintMessage(
            performative=TendermintMessage.Performative.ERROR,  # type: ignore
            error_code=ErrorCode.INVALID_REQUEST,
            error_msg="",
            error_data=dict(message="dummy"),
        ),
    ],
)
def test_serialization(msg: TendermintMessage) -> None:
    """Test the serialization of Tendermint speech-act messages."""

    msg.to = "receiver"
    envelope = Envelope(
        to=msg.to,
        sender="sender",
        message=msg,
    )
    envelope_bytes = envelope.encode()

    actual_envelope = Envelope.decode(envelope_bytes)
    expected_envelope = envelope
    assert expected_envelope.to == actual_envelope.to
    assert expected_envelope.sender == actual_envelope.sender
    assert (
        expected_envelope.protocol_specification_id
        == actual_envelope.protocol_specification_id
    )
    assert expected_envelope.message != actual_envelope.message

    actual_msg = TendermintMessage.serializer.decode(actual_envelope.message)
    actual_msg.to = actual_envelope.to
    actual_msg.sender = actual_envelope.sender
    expected_msg = msg
    assert expected_msg == actual_msg


def test_incorrect_error_data_logged(caplog: LogCaptureFixture) -> None:
    """Test incorrect error_data is logged"""

    expected = "Invalid type for dictionary values in content 'error_data'. Expected 'str'. Found '<class 'int'>'."
    with caplog.at_level(logging.ERROR, logger=_default_logger.name):
        TendermintMessage(
            performative=TendermintMessage.Performative.ERROR,  # type: ignore
            error_code=ErrorCode.INVALID_REQUEST,
            error_msg="",
            error_data=dict(message=1),  # incorrect type
        )
        assert expected in caplog.text


def test_serialization_performative_not_valid() -> None:
    """Test serialization performative not valid"""

    msg = TendermintMessage(
        performative=TendermintMessage.Performative.GET_GENESIS_INFO,  # type: ignore
        query="",
    )
    encoded_msg = msg.encode()

    with mock.patch.object(TendermintMessage, "Performative"):
        with pytest.raises(ValueError, match="Performative not valid: "):
            msg.encode()
        with pytest.raises(ValueError, match="Performative not valid: "):
            msg.decode(encoded_msg)


class AgentDialogue(TendermintDialogue):
    """The dialogue class maintains state of a dialogue and manages it."""

    def __init__(
        self,
        dialogue_label: DialogueLabel,
        self_address: Address,
        role: BaseDialogue.Role,
        message_class: Type[TendermintMessage],
    ) -> None:
        """
        Initialize a dialogue.

        :param dialogue_label: the identifier of the dialogue
        :param self_address: the address of the entity for whom this dialogue is maintained
        :param role: the role of the agent this dialogue is maintained for
        :param message_class: the message class
        """
        TendermintDialogue.__init__(
            self,
            dialogue_label=dialogue_label,
            self_address=self_address,
            role=role,
            message_class=message_class,
        )


class AgentDialogues(TendermintDialogues):
    """The dialogues class keeps track of all dialogues."""

    def __init__(self, self_address: Address) -> None:
        """
        Initialize dialogues.

        :param self_address: the address of the entity for whom this dialogue is maintained
        """

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message,  # pylint: disable=redefined-outer-name
            receiver_address: Address,
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message

            :param message: an incoming/outgoing first message
            :param receiver_address: the address of the receiving agent
            :return: The role of the agent
            """
            return TendermintDialogue.Role.AGENT

        TendermintDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
            dialogue_class=AgentDialogue,
        )


class TestDialogues:
    """Tests abci dialogues."""

    agent_dialogues: AgentDialogues

    @classmethod
    def setup_class(cls) -> None:
        """Setup test class"""

        cls.agent_dialogues = AgentDialogues("agent_address")

    def test_create_self_initiated(self) -> None:
        """Test the self initialisation of a dialogue."""

        result = self.agent_dialogues._create_self_initiated(  # pylint: disable=protected-access
            dialogue_opponent_addr="dummy_address",
            dialogue_reference=(str(0), ""),
            role=TendermintDialogue.Role.AGENT,
        )
        assert isinstance(result, TendermintDialogue)
        assert result.role == TendermintDialogue.Role.AGENT, "The role must be agent."

    def test_create_opponent_initiated(self) -> None:
        """Test the opponent initialisation of a dialogue."""

        result = self.agent_dialogues._create_opponent_initiated(  # pylint: disable=protected-access
            dialogue_opponent_addr="dummy_address",
            dialogue_reference=(str(0), ""),
            role=TendermintDialogue.Role.AGENT,
        )
        assert isinstance(result, TendermintDialogue)
        assert result.role == TendermintDialogue.Role.AGENT, "The role must be agent."



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Generates the specification for a given ABCI app in YAML/JSON/Mermaid format."""

import inspect
import json
import logging
import re
import textwrap
from collections import OrderedDict, defaultdict, deque
from itertools import product
from pathlib import Path
from typing import Any, Dict, List, Set, Tuple

import jsonschema
import yaml
from aea.helpers.io import open_file

from autonomy.configurations.constants import (
    DEFAULT_FSM_SPEC_JSON,
    DEFAULT_FSM_SPEC_MERMAID,
    DEFAULT_FSM_SPEC_YAML,
    SCHEMAS_DIR,
)


FSM_SCHEMA_FILE = "fsm_specification_schema.json"
ROUND_CLASS_POST_FIX = "Round"
ABCI_APP_CLASS_POST_FIX = "AbciApp"

EVENT_PATTERN = re.compile(r"Event\.(\w+)", re.DOTALL)


def validate_fsm_spec(data: Dict) -> None:
    """Validate FSM specificaiton file."""
    with open_file(SCHEMAS_DIR / FSM_SCHEMA_FILE) as fp:
        fsm_schema = json.load(fp=fp)

    validator = jsonschema.Draft4Validator(schema=fsm_schema)
    validator.validate(data)


class DFASpecificationError(Exception):
    """Simple class to raise errors when parsing a DFA."""


class FSMSpecificationLoader:
    """FSM specification loader utilities."""

    class OutputFormats:  # pylint: disable=too-few-public-methods
        """Output formats."""

        JSON = "json"
        YAML = "yaml"
        MERMAID = "mermaid"
        ALL = (JSON, YAML, MERMAID)

        default_output_files = {
            JSON: DEFAULT_FSM_SPEC_JSON,
            YAML: DEFAULT_FSM_SPEC_YAML,
            MERMAID: DEFAULT_FSM_SPEC_MERMAID,
        }

    @staticmethod
    def from_yaml(file: Path) -> Dict:
        """Load from yaml."""
        with open_file(file, mode="r", encoding="utf-8") as fp:
            data = yaml.safe_load(fp)
        return data

    @staticmethod
    def from_json(file: Path) -> Dict:
        """Load from json."""
        with open_file(file, mode="r", encoding="utf-8") as fp:
            data = json.load(fp=fp)
        return data

    @classmethod
    def load(
        cls,
        file: Path,
        spec_format: str = OutputFormats.YAML,
    ) -> Dict:
        """Load FSM specification."""

        if spec_format == cls.OutputFormats.JSON:
            data = cls.from_json(file)
        elif spec_format == cls.OutputFormats.YAML:
            data = cls.from_yaml(file)
        else:
            raise ValueError(f"Unrecognized input format {spec_format}.")

        validate_fsm_spec(data)
        return data

    @staticmethod
    def dump_json(dfa: "DFA", file: Path) -> None:
        """Dump to a json file."""

        with open_file(file, "w", encoding="utf-8") as fp:
            json.dump(dfa.generate(), fp, indent=4)

    @staticmethod
    def dump_yaml(dfa: "DFA", file: Path) -> None:
        """Dump to a yaml file."""
        with open_file(file, "w", encoding="utf-8") as fp:
            yaml.safe_dump(dfa.generate(), fp, indent=4)

    @staticmethod
    def dump_mermaid(dfa: "DFA", file: Path) -> None:
        """Dumps this DFA spec. to a file in Mermaid format."""
        with open_file(file, "w", encoding="utf-8") as fp:
            print("stateDiagram-v2", file=fp)
            aux_map: Dict[Tuple[str, str], Set[str]] = {}
            for (s1, t), s2 in dfa.transition_func.items():
                aux_map.setdefault((s1, s2), set()).add(t)

            # A small optimization to make the output nicer:
            # (1) First, print the arrows that start from a start_state.
            for (s1, s2), t_set in aux_map.items():
                if s1 in dfa.start_states:
                    print(
                        f"    {s1} --> {s2}: <center>{'<br />'.join(t_set)}</center>",
                        file=fp,
                    )

            # (2) Then, print the rest of the arrows.
            for (s1, s2), t_set in aux_map.items():
                if s1 not in dfa.start_states:
                    print(
                        f"    {s1} --> {s2}: <center>{'<br />'.join(t_set)}</center>",
                        file=fp,
                    )

    @classmethod
    def dump(
        cls, dfa: "DFA", file: Path, spec_format: str = OutputFormats.YAML
    ) -> None:
        """Dumps this DFA spec. to a file in YAML/JSON/Mermaid format."""

        validate_fsm_spec(dfa.generate())

        if spec_format == cls.OutputFormats.JSON:
            cls.dump_json(dfa, file)
        elif spec_format == cls.OutputFormats.YAML:
            cls.dump_yaml(dfa, file)
        elif spec_format == cls.OutputFormats.MERMAID:
            cls.dump_mermaid(dfa, file)
        else:
            raise ValueError(f"Unrecognized input format {spec_format}.")


class DFA:
    """Simple specification of a deterministic finite automaton (DFA)."""

    def __init__(
        self,
        label: str,
        states: Set[str],
        default_start_state: str,
        start_states: Set[str],
        final_states: Set[str],
        alphabet_in: Set[str],
        transition_func: Dict[Tuple[str, str], str],
    ):  # pylint: disable=too-many-arguments
        """Initialize DFA object."""
        transition_func_in_states, transition_func_alphabet_in = map(
            set, zip(*transition_func.keys())
        )

        transition_func_states = transition_func_in_states.copy()
        transition_func_states.update(transition_func.values())  # type: ignore

        reachable_states = DFA._get_reachable_states(start_states, transition_func)
        unreachable_states = states - reachable_states

        error_strings = []
        if unreachable_states:
            error_strings.append(
                f" - DFA spec. contains unreachable states: {unreachable_states}."
            )
        if not transition_func_states.issubset(states):
            error_strings.append(
                f" - Transition function contains unexpected states: {transition_func_states - states}."  # type: ignore
            )
        if not transition_func_alphabet_in.issubset(alphabet_in):
            error_strings.append(
                " - Transition function contains unexpected input symbols: "
                f"{transition_func_alphabet_in - alphabet_in}."  # type: ignore
            )
        if not alphabet_in.issubset(transition_func_alphabet_in):
            error_strings.append(
                f" - Unused input symbols: {alphabet_in - transition_func_alphabet_in}."  # type: ignore
            )
        if default_start_state not in start_states:
            error_strings.append(" - Default start state is not in start states set.")
        if not start_states.issubset(states):
            error_strings.append(
                f" - Start state set contains unexpected states: {start_states - states}."
            )
        if not final_states.issubset(states):
            error_strings.append(
                f" - Final state set contains unexpected states: {final_states - states}."
            )
        if start_states & final_states:
            error_strings.append(
                f" - Final state set contains start states: {start_states & final_states}."
            )
        if transition_func_in_states & final_states:
            error_strings.append(
                f" - Transitions out from final states: {transition_func_in_states & final_states}."
            )

        if len(error_strings) > 0:
            raise DFASpecificationError(
                "DFA spec has the following issues:\n" + "\n".join(error_strings)
            )

        self.label = label
        self.states = states
        self.default_start_state = default_start_state
        self.start_states = start_states
        self.final_states = final_states
        self.alphabet_in = alphabet_in
        self.transition_func = transition_func

        self.validate_naming_conventions()

    def validate_naming_conventions(self) -> None:
        """
        Validate state names to see if they follow the naming conventions below

        - A round name should end with `Round`
        - ABCI app class name should end with `AbciApp`
        """

        if not self.label.endswith(ABCI_APP_CLASS_POST_FIX):
            raise DFASpecificationError(
                f"ABCI app class name should end in `AbciApp`; ABCI app name found `{self.label}`"
            )

        for state in self.states:
            if not state.endswith(ROUND_CLASS_POST_FIX):
                raise DFASpecificationError(
                    f"Round class name should end in `Round`; Round app name found `{state}`"
                )

    def is_transition_func_total(self) -> bool:
        """
        Outputs True if the transition function of the DFA is total.

        A transition function is total when it explicitly defines all the transitions
        for all the possible pairs (state, input_symbol). By convention, when a transition
        (state, input_symbol) is not defined for a certain input_symbol, it will be
        automatically regarded as a self-transition to the same state.

        :return: True if the transition function is total. False otherwise.
        """
        return set(product(self.states, self.alphabet_in)) == set(
            self.transition_func.keys()
        )

    def get_transitions(self, input_sequence: List[str]) -> List[str]:
        """Runs the DFA given the input sequence of symbols, and outputs the list of state transitions."""
        state = self.default_start_state
        transitions = [state]
        for t in input_sequence:
            if t not in self.alphabet_in:
                logging.warning(
                    f"Input sequence contains a symbol {t} (ignored) "
                    f"not belonging to the DFA alphabet {self.alphabet_in}."
                )
            else:
                state = self.transition_func.get((state, t), state)
                transitions.append(state)
        return transitions

    def parse_transition_func(self) -> Dict[str, Dict[str, str]]:
        """Parse the transition function from the spec to a nested dictionary."""

        result: Dict[str, Dict[str, str]] = {}
        for (round_cls_name, event_name), value in self.transition_func.items():
            result.setdefault(round_cls_name, {})[f"Event.{event_name}"] = value
        for state in self.states:
            if state not in result:
                result[state] = {}
        return result

    def __eq__(self, other: object) -> bool:
        """Compares two DFAs"""
        if not isinstance(other, DFA):
            return NotImplemented  # Try reflected operation
        return self.__dict__ == other.__dict__

    def generate(self) -> Dict[str, Any]:
        """Retrieves an exportable representation for YAML/JSON dump of this DFA."""
        dfa_export: Dict[str, Any] = {}
        for k, v in self.__dict__.items():
            if isinstance(v, Set):
                dfa_export[k] = sorted(v)
            elif isinstance(v, Dict):
                dfa_export[k] = dict(
                    OrderedDict(
                        [(str(k2).replace("'", ""), v2) for k2, v2 in v.items()]
                    )
                )
            else:
                dfa_export[k] = v
        return dfa_export

    @classmethod
    def _get_reachable_states(
        cls,
        start_states: Set[str],
        transition_func: Dict[Tuple[str, str], str],
    ) -> Set[str]:
        """Performs a BFS-like search returning the reachable set of states departing from the set of start states."""

        graph: Dict[str, Set[str]] = defaultdict(set)
        for (s1, _), s2 in transition_func.items():
            graph[s1].add(s2)

        visited: Set[str] = start_states.copy()
        queue = deque(start_states)

        while queue:
            s = queue.popleft()

            for neighbour in graph[s]:
                if neighbour not in visited:
                    visited.add(neighbour)
                    queue.append(neighbour)

        return visited

    @classmethod
    def _norep_list_to_set(cls, lst: List[str]) -> Set[str]:
        """Converts a list to a set and ensures that it does not contain repetitions."""
        if not isinstance(lst, List):
            raise DFASpecificationError(f"DFA spec. object {lst} is not of type List.")
        if len(lst) != len(set(lst)):
            raise DFASpecificationError(
                f"DFA spec. List {lst} contains repeated values."
            )
        return set(lst)

    @classmethod
    def _str_to_tuple(cls, k: str) -> Tuple[str, str]:
        """Converts a string in format "(a, b)" to a tuple ("a", "b")."""
        match = re.search(r"\((\w+),\s*(\w+)\)", k, re.DOTALL)

        if match is None:
            raise DFASpecificationError(
                f"DFA spec. JSON file contains an invalid transition function key: {k}."
            )

        return match.group(1), match.group(2)

    @classmethod
    def load(
        cls, file: Path, spec_format: str = FSMSpecificationLoader.OutputFormats.YAML
    ) -> "DFA":
        """Loads a DFA JSON specification from file."""

        dfa_data = FSMSpecificationLoader.load(file=file, spec_format=spec_format)

        label = dfa_data.pop("label")
        states = DFA._norep_list_to_set(dfa_data.pop("states"))
        default_start_state = dfa_data.pop("default_start_state")
        start_states = DFA._norep_list_to_set(dfa_data.pop("start_states"))
        final_states = DFA._norep_list_to_set(dfa_data.pop("final_states"))
        alphabet_in = DFA._norep_list_to_set(dfa_data.pop("alphabet_in"))
        transition_func = {
            DFA._str_to_tuple(k): v for k, v in dfa_data.pop("transition_func").items()
        }

        return cls(
            label,
            states,
            default_start_state,
            start_states,
            final_states,
            alphabet_in,
            transition_func,
        )

    @classmethod
    def abci_to_dfa(cls, abci_app_cls: Any, label: str = "") -> "DFA":
        """Translates an AbciApp class into a DFA."""

        trf = abci_app_cls.transition_function

        label = label if label != "" else abci_app_cls.__name__
        default_start_state = abci_app_cls.initial_round_cls.__name__
        start_states = DFA._norep_list_to_set(
            [s.__name__ for s in abci_app_cls.initial_states]
        )
        start_states = start_states if start_states else {default_start_state}
        final_states = DFA._norep_list_to_set(
            [s.__name__ for s in abci_app_cls.final_states]
        )
        states = (
            DFA._norep_list_to_set([k.__name__ for k in trf])
            | {s.__name__ for k in trf for s in trf[k].values()}
            | start_states
            | final_states
        )
        alphabet_in = {str(s).rsplit(".", 1)[1] for k in trf for s in trf[k].keys()}
        transition_func = {
            (k.__name__, str(s).rsplit(".", 1)[1]): trf[k][s].__name__
            for k in trf
            for s in trf[k]
        }
        transition_func = OrderedDict(sorted(transition_func.items()))
        return cls(
            label,
            states,
            default_start_state,
            start_states,
            final_states,
            alphabet_in,
            transition_func,
        )


def check_unreferenced_events(abci_app_cls: Any) -> List[str]:
    """Checks for unreferenced events in the AbciApp.

    Checks that events defined in the AbciApp transition function are referenced
    in the source code of the corresponding rounds or their superclasses. Note that
    the function simply checks references in the "raw" source code of the rounds and
    their (non builtin) superclasses. Therefore, it does not do any kind of static
    analysis on the source code, nor checks for actual reachability of a return
    statement returning such events.

    :param abci_app_cls: AbciApp to check unreferenced events.
    :return: List of error strings
    """

    error_strings = []
    timeout_events = {k.name for k in abci_app_cls.event_to_timeout.keys()}

    for round_cls, round_transitions in abci_app_cls.transition_function.items():
        trf_events = {
            str(e).rsplit(".", 1)[1] for e in round_transitions
        } - timeout_events
        referenced_events = set()

        for base in filter(
            lambda x: x.__class__.__module__ != "builtins",
            inspect.getmro(round_cls),
        ):
            src = textwrap.dedent(inspect.getsource(base))
            referenced_events.update(EVENT_PATTERN.findall(src))

        if trf_events.symmetric_difference(referenced_events):
            error_strings.append(
                f" - {round_cls.__name__}: transition function events {trf_events} "
                f"do not match referenced events {referenced_events}."
            )

    return error_strings



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test for contract module."""

from pathlib import Path

from aea_test_autonomy.base_test_classes.contracts import BaseRegistriesContractsTest
from aea_test_autonomy.docker.base import skip_docker_tests
from aea_test_autonomy.docker.registries import SERVICE_MANAGER

from packages.valory.contracts.service_manager.contract import (
    JSONLike,
    ServiceManagerContract,
)


PACKAGE_DIR = Path(__file__).parent.parent

METADATA_HASH = "0xaaca27f6156089376fd85900d511c7570b6c9b0f6afc64576aac5aa0da8de92d"
AGENT_ID = 1
NUMBER_OF_SLOTS = 1
COST_OF_BOND = 1000
THRESHOLD = 3


@skip_docker_tests
class TestServiceManager(BaseRegistriesContractsTest):
    """Test service manager."""

    contract: ServiceManagerContract
    contract_address = SERVICE_MANAGER
    contract_directory = PACKAGE_DIR

    def make_transaction(self, tx: JSONLike) -> JSONLike:
        """Make transaction."""
        tx_signed = self.deployer_crypto.sign_transaction(transaction=tx)
        tx_digest = self.ledger_api.send_signed_transaction(tx_signed=tx_signed)

        return self.ledger_api.get_transaction_receipt(tx_digest=tx_digest)

    def test_get_create_transaction(self) -> None:
        """Test `get_create_transaction` method."""
        tx = self.contract.get_create_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            owner=self.deployer_crypto.address,
            sender=self.deployer_crypto.address,
            metadata_hash=METADATA_HASH,
            agent_ids=[AGENT_ID],
            agent_params=[[NUMBER_OF_SLOTS, COST_OF_BOND]],
            threshold=THRESHOLD,
        )
        assert all(
            [
                key
                in [
                    "chainId",
                    "nonce",
                    "value",
                    "gas",
                    "maxFeePerGas",
                    "maxPriorityFeePerGas",
                    "to",
                    "data",
                ]
                for key in tx.keys()
            ]
        )

    def test_get_activate_registration_transaction(self) -> None:
        """Test `get_activate_registration_transaction` method"""

        tx = self.contract.get_activate_registration_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            owner=self.deployer_crypto.address,
            service_id=1,
            security_deposit=COST_OF_BOND,
        )
        assert all(
            [
                key
                in [
                    "chainId",
                    "nonce",
                    "value",
                    "gas",
                    "maxFeePerGas",
                    "maxPriorityFeePerGas",
                    "to",
                    "data",
                ]
                for key in tx.keys()
            ]
        )

    def test_get_terminate_service_transaction(self) -> None:
        """Test `get_terminate_service_transaction` method"""

        tx = self.contract.get_terminate_service_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            owner=self.deployer_crypto.address,
            service_id=1,
        )
        assert all(
            [
                key
                in [
                    "chainId",
                    "nonce",
                    "value",
                    "gas",
                    "maxFeePerGas",
                    "maxPriorityFeePerGas",
                    "to",
                    "data",
                ]
                for key in tx.keys()
            ]
        )

    def test_get_unbond_service_transaction(self) -> None:
        """Test `get_unbond_service_transaction` method"""

        tx = self.contract.get_unbond_service_transaction(
            ledger_api=self.ledger_api,
            contract_address=self.contract_address,
            owner=self.deployer_crypto.address,
            service_id=1,
        )
        assert all(
            [
                key
                in [
                    "chainId",
                    "nonce",
                    "value",
                    "gas",
                    "maxFeePerGas",
                    "maxPriorityFeePerGas",
                    "to",
                    "data",
                ]
                for key in tx.keys()
            ]
        )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for the Tendermint com server."""

import json
import logging
import os
import platform
import shutil
import stat
import subprocess  # nosec
import tempfile
import time
from pathlib import Path
from typing import Callable, Dict, Set, cast
from unittest import mock

import flask
import pytest
import requests
from _pytest.logging import LogCaptureFixture  # type: ignore
from _pytest.monkeypatch import MonkeyPatch  # type: ignore
from aea.common import JSONLike
from aea.test_tools.utils import wait_for_condition

from deployments.Dockerfiles.tendermint import app  # type: ignore
from deployments.Dockerfiles.tendermint.app import TendermintNode  # type: ignore
from deployments.Dockerfiles.tendermint.app import (  # type: ignore
    CONFIG_OVERRIDE,
    DOCKER_INTERNAL_HOST,
    PeriodDumper,
    create_app,
    get_defaults,
    load_genesis,
    override_config_toml,
    update_peers,
)
from deployments.Dockerfiles.tendermint.tendermint import (  # type: ignore
    TendermintParams,
)


ENCODING = "utf-8"
VERSION = "0.34.19"
DUMMY_CONFIG_TOML = """
# Comma separated list of seed nodes to connect to
seeds = ""

# Comma separated list of nodes to keep persistent connections to
persistent_peers = "dummy_peer1@localhost:80,dummy_peer2@localhost:81"

# UPNP port forwarding
upnp = false

external_address = ""
"""
DUMMY_EXTERNAL_ADDRESS = "dummy_external_address:26566"

wait_for_node_to_run = pytest.mark.usefixtures("wait_for_node")
wait_for_occupied_rpc_port = pytest.mark.usefixtures("wait_for_occupied_rpc_port")


# utility functions
def readonly_handler(func: Callable, path: str, execinfo) -> None:  # type: ignore
    """If permission is readonly, we change and retry."""
    os.chmod(path, stat.S_IWRITE)
    func(path)


def test_period_dumper(monkeypatch: MonkeyPatch) -> None:
    """Test PeriodDumper"""

    monkeypatch.setenv("ID", "42")
    with tempfile.TemporaryDirectory() as tmp_dir:
        monkeypatch.setenv("TMHOME", tmp_dir)
        dump_dir = Path(tempfile.mkdtemp())
        period_dumper = PeriodDumper(mock.Mock(), dump_dir=dump_dir)
        period_dumper.dump_period()

    shutil.rmtree(period_dumper.dump_dir)


# base classes
class BaseTendermintTest:
    """BaseTendermintTest"""

    tendermint: str
    tm_home: str
    path: Path
    _os_env: Dict

    @classmethod
    def setup_class(cls) -> None:
        """Setup the test."""
        cls._os_env = os.environ.copy()
        cls.tendermint = shutil.which("tendermint")  # type: ignore
        cls.tm_home = os.environ["TMHOME"] = tempfile.mkdtemp()
        cls.path = Path(cls.tm_home)
        assert not os.listdir(cls.tm_home)

        command = [cls.tendermint, "init", "validator", "--home", cls.tm_home]
        process = subprocess.Popen(command, stderr=subprocess.PIPE)  # nosec
        _, stderr = process.communicate()
        assert not stderr, stderr

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown the test."""
        os.environ.clear()
        os.environ.update(cls._os_env)
        shutil.rmtree(cls.tm_home, ignore_errors=True, onerror=readonly_handler)


class BaseTendermintServerTest(BaseTendermintTest):
    """Test Tendermint server app"""

    app: flask.app.Flask
    app_context: flask.ctx.AppContext
    tendermint_node: TendermintNode
    perform_monitoring = True
    debug_tendermint = False
    tm_status_endpoint = "http://localhost:26657/status"
    dump_dir: Path

    @classmethod
    def setup_class(cls) -> None:
        """Setup the test."""
        super().setup_class()
        os.environ["PROXY_APP"] = "kvstore"
        os.environ["CREATE_EMPTY_BLOCKS"] = "true"
        os.environ["USE_GRPC"] = "false"
        os.environ["LOG_FILE"] = str(cls.path / "tendermint.log")
        cls.dump_dir = Path(tempfile.mkdtemp())
        cls.app, cls.tendermint_node = create_app(
            dump_dir=cls.dump_dir,
            perform_monitoring=cls.perform_monitoring,
            debug=cls.debug_tendermint,
        )
        cls.app.config["TESTING"] = True
        cls.app_context = cls.app.app_context()
        cls.app_context.push()

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown the test."""
        cls.app_context.pop()
        cls.tendermint_node.stop()
        shutil.rmtree(cls.dump_dir)
        super().teardown_class()


# unit tests
def test_tendermint_executable_found() -> None:
    """Test Tendermint executable found"""
    assert shutil.which("tendermint"), "No `tendermint` executable found"
    output = subprocess.check_output(["tendermint", "version"])  # nosec
    assert output.decode(ENCODING).strip() == VERSION


class TestTendermintServerUtilityFunctions(BaseTendermintTest):
    """Test Tendermint server utility functions"""

    def test_load_genesis(self) -> None:
        """Load the genesis.json file"""
        expected_keys = {
            "genesis_time",
            "chain_id",
            "initial_height",
            "consensus_params",
            "validators",
            "app_hash",
        }
        json = load_genesis()
        assert set(json.keys()) == expected_keys

    def test_get_defaults(self) -> None:
        """Test get_defaults function"""
        expected_keys = {"genesis_time"}
        defaults = get_defaults()
        assert set(defaults.keys()) == expected_keys

    def test_override_config_toml(self) -> None:
        """Test override_config_toml function"""
        path = self.path / "config" / "config.toml"
        assert path.is_file(), f"not a file: {path}"
        txt = path.read_text("utf-8")
        assert all(old in txt and new not in txt for old, new in CONFIG_OVERRIDE)
        override_config_toml()
        txt = path.read_text("utf-8")
        assert not any(old in txt or new not in txt for old, new in CONFIG_OVERRIDE)


# integration tests
class TestTendermintServerApp(BaseTendermintServerTest):
    """Test Tendermint server app"""

    genesis_filepath: Path
    genesis_config: Dict

    def setup(self) -> None:
        """Setup"""
        self.genesis_filepath = Path(os.environ["TMHOME"], "config", "genesis.json")
        self.genesis_config = load_genesis()

    def teardown(self) -> None:
        """Teardown"""
        self.genesis_filepath.write_text(json.dumps(self.genesis_config))

    @wait_for_node_to_run
    def test_files_exist(self) -> None:
        """Test that the necessary files are present"""

        expected_file_names = [
            Path(self.tm_home, "config", "config.toml"),
            Path(self.tm_home, "config", "priv_validator_key.json"),
            Path(self.tm_home, "config", "genesis.json"),
            Path(self.tm_home, "config", "node_key.json"),
            Path(self.tm_home, "data", "priv_validator_state.json"),
        ]

        assert all(f.exists() for f in expected_file_names), expected_file_names

    @wait_for_node_to_run
    def test_get_request_status(self, http_: str, loopback: str, rpc_port: int) -> None:
        """Check local node is running"""
        response = requests.get(f"{http_}{loopback}:{rpc_port}/status")
        data = response.json()
        assert data["result"]["node_info"]["version"] == VERSION

    @wait_for_node_to_run
    def test_handle_notfound(self) -> None:
        """Test handle not found"""
        with self.app.test_client() as client:
            response = client.get("/non_existing_endpoint")
            assert response.status_code == 404

    @wait_for_node_to_run
    def test_get_app_hash(self) -> None:
        """Test get app hash"""
        with self.app.test_client() as client:
            wait_for_condition(
                lambda: "app_hash" in client.get("/app_hash").get_json(),
                timeout=10,
                period=1,
                error_msg="Could not retrieve `app_hash` from Tendermint server",
            )

    @wait_for_node_to_run
    def test_get_params(self) -> None:
        """Test get app hash"""

        with self.app.test_client() as client:
            response = client.get("/params")
            assert response.status_code == 200
            data = cast(JSONLike, response.get_json())
            assert data["status"] is True
            assert data["error"] is None
            params = data.get("params")
            assert params
            assert params.pop("address")
            assert params.pop("peer_id")
            pub_key = params.pop("pub_key")
            assert pub_key.pop("type")
            assert pub_key.pop("value")
            # nothing else should be sent, certainly not a private key.
            assert not params

    @wait_for_node_to_run
    def test_update_params(self) -> None:
        """Test get app hash"""

        dummy_data = dict(
            genesis_config=load_genesis(),
            validators=[],
            external_address="",
        )

        with self.app.test_client() as client:
            response = client.post("/params", json=dummy_data)
            assert response.status_code == 200
            data = cast(JSONLike, response.get_json())
            assert data["status"] is True
            assert data["error"] is None

    @wait_for_node_to_run
    def test_get_and_update(self) -> None:
        """Test get and update Tendermint genesis data"""

        genesis_config = load_genesis()

        with self.app.test_client() as client:
            # 1. check params are in validator set
            response = client.get("/params")
            assert response.status_code == 200
            response_data = cast(Dict, response.get_json())
            params = cast(Dict, response_data["params"])
            params.pop("peer_id", None)
            params["name"], params["power"] = "", "10"
            assert params in genesis_config["validators"]

            # 2. clear validators
            dummy_data = dict(
                genesis_config=genesis_config,
                validators=[],
                external_address=DUMMY_EXTERNAL_ADDRESS,
            )
            response = client.post("/params", json=dummy_data)
            assert response.status_code == 200
            genesis_config = load_genesis()
            assert params not in genesis_config["validators"]


class TestTendermintGentleResetServer(BaseTendermintServerTest):
    """Test Tendermint gentle reset"""

    @wait_for_node_to_run
    def test_gentle_reset(self) -> None:
        """Test gentle reset"""
        with self.app.test_client() as client:
            response = client.get("/gentle_reset")
            data = response.get_json()
            assert response.status_code == 200
            assert data["status"] is True


class TestTendermintHardResetServer(BaseTendermintServerTest):
    """Test Tendermint hard reset"""

    @wait_for_node_to_run
    @pytest.mark.parametrize("prune_fail", (True, False))
    def test_hard_reset(self, prune_fail: bool) -> None:
        """Test hard reset"""
        with self.app.test_client() as client, mock.patch.object(
            TendermintNode, "prune_blocks", return_value=int(prune_fail)
        ):
            response = client.get("/hard_reset")
            data = response.get_json()
            assert response.status_code == 200
            assert data["status"] is not prune_fail

    @wait_for_node_to_run
    @pytest.mark.skipif(platform.system() != "Linux", reason="scoped to docker image")
    def test_hard_reset_dev_mode(self, caplog: LogCaptureFixture) -> None:
        """Test hard reset"""

        resets = 0  # zero since first hard reset
        os.environ["ID"] = "_dummy_ID"
        period_dumper = PeriodDumper(logger=logging.getLogger(), dump_dir=self.dump_dir)
        store_dir = period_dumper.dump_dir / f"period_{resets}"
        path = store_dir / ("node" + os.environ["ID"])
        logging.error(f"expected: {path}")
        assert not path.exists()

        with self.app.test_client() as client:
            with mock.patch.object(app, "IS_DEV_MODE", return_value=True):
                response = client.get("/hard_reset")
                assert response.status_code == 200
                data = cast(JSONLike, response.get_json())
                assert data["status"] is True
                assert "Dumped data for period" in caplog.text

        assert path.exists()
        expected = {"config", "tendermint.log", "data"}
        assert {p.name for p in path.glob("*")} == expected


class TestTendermintLogMessages(BaseTendermintServerTest):
    """Test Tendermint message logging"""

    @wait_for_node_to_run
    def test_tendermint_logs(self) -> None:
        """Test Tendermint logs"""

        def get_logs() -> str:
            with open(os.environ["LOG_FILE"], "r") as f:
                lines = "".join(f.readlines())
            return lines

        def get_missing(messages: Set[str]) -> Set[str]:
            while messages:
                logs = get_logs()
                messages -= {line for line in messages if line in logs}
            return messages

        before_stopping = {
            "Tendermint process started",
            "Monitoring thread started",
            "Starting multiAppConn service",
            "Starting localClient service",
            "This node is a validator",
        }

        after_stopping = {
            "Monitoring thread terminated",
            "Tendermint process stopped",
        }

        wait_for_condition(
            lambda: not get_missing(before_stopping),
            error_msg=f"Not all `before_stopping` messages found in Tendermint logs: {before_stopping}",
            timeout=10,
            period=1,
        )
        self.tendermint_node.stop()
        wait_for_condition(
            lambda: not get_missing(after_stopping),
            error_msg=f"Not all `after_stopping` messages found in Tendermint logs: {after_stopping}",
            timeout=10,
            period=1,
        )


def mock_get_node_command_kwargs() -> Dict:
    """Get the node command kwargs"""
    kwargs = {
        "bufsize": 1,
        "universal_newlines": True,
    }

    # Pipe stdout and stderr even if we're not going to read to provoke the buffer fill
    kwargs["stdout"] = subprocess.PIPE
    kwargs["stderr"] = subprocess.STDOUT

    if platform.system() == "Windows":  # pragma: nocover
        kwargs["creationflags"] = subprocess.CREATE_NEW_PROCESS_GROUP  # type: ignore
    else:
        kwargs["preexec_fn"] = os.setsid  # type: ignore

    return kwargs


class TestTendermintBufferFailing(BaseTendermintServerTest):
    """Test Tendermint buffer"""

    perform_monitoring = False  # Setting this flag to False and not monitoring makes the buffer to fill and freeze Tendermint
    debug_tendermint = True  # This will cause more logging -> faster failure

    @classmethod
    def setup_class(cls) -> None:
        """Setup the test."""
        with mock.patch.object(
            TendermintParams,
            "get_node_command_kwargs",
            return_value=mock_get_node_command_kwargs(),
        ):
            super().setup_class()

    @wait_for_occupied_rpc_port
    def test_tendermint_buffer(self) -> None:
        """Test Tendermint buffer"""

        with pytest.raises(requests.exceptions.Timeout):
            # Give the test 30 seconds for it to throw a timeout
            for _ in range(30):
                # If it hangs for 5 seconds, we assume it's not working.
                # Increasing the timeout should have no effect,
                # the node is unresponsive at this point.
                requests.get(self.tm_status_endpoint, timeout=5)
                time.sleep(1)

    @classmethod
    def teardown_class(cls) -> None:
        """Teardown the test."""
        # After the test, the node has hanged. We need to kill it and not stop it.
        if cls.tendermint_node._process:
            cls.tendermint_node._process.kill()
        cls.app_context.pop()
        shutil.rmtree(cls.tm_home, ignore_errors=True, onerror=readonly_handler)


class TestTendermintBufferWorking(BaseTendermintServerTest):
    """Test Tendermint buffer"""

    perform_monitoring = True
    debug_tendermint = True

    @wait_for_node_to_run
    def test_tendermint_buffer(self) -> None:
        """Test Tendermint buffer"""

        # Give the test 60 seconds for it to work
        for _ in range(60):
            try:
                res = requests.get(self.tm_status_endpoint, timeout=5)
                # We expect all responses to be OK
                assert res.status_code == 200
            except Exception as e:
                raise AssertionError(e)

            time.sleep(1)


def test_update_peers() -> None:
    """Test update peers."""

    validators = [
        {
            "hostname": f"host_{i}",
            "p2p_port": i * 10,
            "peer_id": f"peer_{i}",
        }
        for i in range(2)
    ]

    with tempfile.TemporaryDirectory() as temp_dir:
        config_file = Path(temp_dir, "config.toml")
        config_file.write_text(DUMMY_CONFIG_TOML)

        update_peers(validators=validators, config_path=config_file)

        updated_content = config_file.read_text()

        # make sure we don't update anything else
        assert """seeds = """ "" in updated_content
        assert "upnp = false" in updated_content

        for val in validators:
            assert (
                str(val["peer_id"])
                + "@"
                + str(val["hostname"])
                + ":"
                + str(val["p2p_port"])
                in updated_content
            )


def test_update_peers_internal_host() -> None:
    """Test update peers."""

    validators = [{"hostname": "localhost", "p2p_port": 80, "peer_id": "peer"}]

    with tempfile.TemporaryDirectory() as temp_dir:
        config_file = Path(temp_dir, "config.toml")
        config_file.write_text(DUMMY_CONFIG_TOML)

        update_peers(validators=validators, config_path=config_file)

        updated_content = config_file.read_text()

        # make sure we don't update anything else
        assert """seeds = """ "" in updated_content
        assert "upnp = false" in updated_content

        for val in validators:
            assert (
                str(val["peer_id"])
                + "@"
                + DOCKER_INTERNAL_HOST
                + ":"
                + str(val["p2p_port"])
                in updated_content
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Base deployments module."""
import abc
import json
import logging
import os
from copy import copy, deepcopy
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple
from warnings import warn

from aea.configurations.base import (
    ConnectionConfig,
    ContractConfig,
    ProtocolConfig,
    SkillConfig,
)
from aea.configurations.constants import SKILL
from aea.configurations.data_types import PackageType, PublicId
from aea.helpers.env_vars import apply_env_variables

from autonomy.analyse.service import ABCI
from autonomy.configurations.base import Service
from autonomy.configurations.loader import load_service_config
from autonomy.constants import DEFAULT_DOCKER_IMAGE_AUTHOR
from autonomy.deploy.constants import (
    DEFAULT_ENCODING,
    INFO,
    KEY_SCHEMA_ADDRESS,
    KEY_SCHEMA_PRIVATE_KEY,
    KEY_SCHEMA_TYPE,
)


ENV_VAR_ID = "ID"
ENV_VAR_AEA_AGENT = "AEA_AGENT"
ENV_VAR_LOG_LEVEL = "LOG_LEVEL"
ENV_VAR_AEA_PASSWORD = "AEA_PASSWORD"  # nosec

PARAM_ARGS_PATH = ("models", "params", "args")
SETUP_PARAM_PATH = (*PARAM_ARGS_PATH, "setup")
SAFE_CONTRACT_ADDRESS = "safe_contract_address"
ALL_PARTICIPANTS = "all_participants"
CONSENSUS_THRESHOLD = "consensus_threshold"


DEFAULT_ABCI_PORT = 26658


KUBERNETES_DEPLOYMENT = "kubernetes"
DOCKER_COMPOSE_DEPLOYMENT = "docker-compose"

LOCALHOST = "localhost"
TENDERMINT_P2P_PORT = 26656

TENDERMINT_NODE = "http://{host}:26657"
TENDERMINT_COM = "http://{host}:8080"

TENDERMINT_NODE_LOCAL = f"http://{LOCALHOST}:26657"
TENDERMINT_COM_LOCAL = f"http://{LOCALHOST}:8080"

TENDERMINT_URL_PARAM = "tendermint_url"
TENDERMINT_COM_URL_PARAM = "tendermint_com_url"

TENDERMINT_P2P_URL = "{host}:{port}"
TENDERMINT_P2P_URL_PARAM = "tendermint_p2p_url"
TENDERMINT_P2P_URL_ENV_VAR = "TM_P2P_NODE_URL_{}"

COMPONENT_CONFIGS: Dict = {
    component.package_type.value: component  # type: ignore
    for component in [
        ContractConfig,
        SkillConfig,
        ProtocolConfig,
        ConnectionConfig,
    ]
}


class NotValidKeysFile(Exception):
    """Raise when provided keys file is not valid."""


class ServiceBuilder:  # pylint: disable=too-many-instance-attributes
    """Class to assist with generating deployments."""

    deplopyment_type: str = DOCKER_COMPOSE_DEPLOYMENT
    log_level: str = INFO

    def __init__(  # pylint: disable=too-many-arguments
        self,
        service: Service,
        keys: Optional[List[Dict[str, str]]] = None,
        private_keys_password: Optional[str] = None,
        agent_instances: Optional[List[str]] = None,
        apply_environment_variables: bool = False,
    ) -> None:
        """Initialize the Base Deployment."""
        if apply_environment_variables:
            warn(  # pragma: no cover
                "`apply_environment_variables` argument is deprecated and will be removed in v1.0.0, "
                "usage of environment varibales is default now.",
                DeprecationWarning,
                stacklevel=2,
            )

        self.service = service

        self._service_name_clean = self.service.name.replace("_", "")
        self._keys = keys or []
        self._agent_instances = agent_instances
        self._private_keys_password = private_keys_password
        self._all_participants = self.try_get_all_participants()

    def get_abci_container_name(self, index: int) -> str:
        """Format ABCI container name."""
        return f"{self._service_name_clean}_abci_{index}"

    def get_tm_container_name(self, index: int) -> str:
        """Format tendermint container name."""
        return f"{self._service_name_clean}_tm_{index}"

    def try_get_all_participants(self) -> Optional[List[str]]:
        """Try get all participants from the ABCI overrides"""
        try:
            for override in deepcopy(self.service.overrides):
                (
                    override,
                    component_id,
                    has_multiple_overrides,
                ) = self.service.process_metadata(
                    configuration=override,
                )
                if (
                    component_id.component_type.value == SKILL
                    and has_multiple_overrides
                ):
                    override, *_ = override.values()

                override = apply_env_variables(
                    data=override,
                    env_variables=os.environ.copy(),
                )
                setup_param = self._get_config_from_json_path(
                    override_dict=override, json_path=SETUP_PARAM_PATH
                )
                return setup_param.get(ALL_PARTICIPANTS)
        except KeyError:
            return None

        return None

    @property
    def private_keys_password(
        self,
    ) -> Optional[str]:
        """Service password for agent keys."""

        password = self._private_keys_password
        if password is None:
            password = os.environ.get("AUTONOLAS_SERVICE_PASSWORD")

        return password

    @property
    def agent_instances(
        self,
    ) -> Optional[List[str]]:
        """Agent instances."""

        return self._agent_instances

    @agent_instances.setter
    def agent_instances(self, instances: List[str]) -> None:
        """Agent instances setter."""

        if self.keys:
            self.verify_agent_instances(
                keys=self.keys,
                agent_instances=instances,
            )

        self._agent_instances = instances

    @property
    def keys(
        self,
    ) -> List[Dict[str, str]]:
        """Keys."""
        return self._keys

    @classmethod
    def from_dir(  # pylint: disable=too-many-arguments
        cls,
        path: Path,
        keys_file: Optional[Path] = None,
        number_of_agents: Optional[int] = None,
        private_keys_password: Optional[str] = None,
        agent_instances: Optional[List[str]] = None,
        apply_environment_variables: bool = False,
    ) -> "ServiceBuilder":
        """Service builder from path."""
        service = load_service_config(
            service_path=path,
        )
        if number_of_agents is not None:
            service.number_of_agents = number_of_agents

        service_builder = cls(
            service=service,
            apply_environment_variables=apply_environment_variables,
            private_keys_password=private_keys_password,
        )

        if keys_file is not None:
            service_builder.read_keys(keys_file=keys_file)

        if agent_instances is not None:
            service_builder.agent_instances = agent_instances

        return service_builder

    @staticmethod
    def verify_agent_instances(
        keys: List[Dict[str, str]],
        agent_instances: List[str],
    ) -> None:
        """Cross verify agent instances with the keys."""
        addresses = {kp["address"] for kp in keys}
        instances = set(agent_instances)

        key_not_in_instances = addresses.difference(instances)
        if key_not_in_instances:
            raise NotValidKeysFile(
                f"Key file contains keys which are not registered as instances; invalid keys={key_not_in_instances}"
            )

        instances_not_in_keys = instances.difference(addresses)
        if instances_not_in_keys:
            logging.warning(
                f"Key file does not contain key pair for following instances {instances_not_in_keys}"
            )

        keys_found_with_instances = instances.intersection(addresses)
        logging.info(
            f"Found following keys with registered instances {keys_found_with_instances}"
        )

    def read_keys(self, keys_file: Path) -> None:
        """Read in keys from a file on disk."""

        try:
            keys = json.loads(keys_file.read_text(encoding=DEFAULT_ENCODING))
        except json.decoder.JSONDecodeError as e:  # pragma: nocover
            raise NotValidKeysFile(
                "Error decoding keys file, please check the content of the file"
            ) from e

        for key in keys:
            if {KEY_SCHEMA_ADDRESS, KEY_SCHEMA_PRIVATE_KEY} != set(key.keys()) and {
                KEY_SCHEMA_ADDRESS,
                KEY_SCHEMA_PRIVATE_KEY,
                KEY_SCHEMA_TYPE,
            } != set(key.keys()):
                raise NotValidKeysFile("Key file incorrectly formatted.")

        if self.agent_instances is not None:
            self.verify_agent_instances(
                keys=keys,
                agent_instances=self.agent_instances,
            )
            self.service.number_of_agents = len(keys)

        if self._all_participants is not None and len(self._all_participants) > 0:
            unwanted_keys = set(key["address"] for key in keys) - set(
                self._all_participants
            )
            if len(unwanted_keys) > 0:
                raise NotValidKeysFile(
                    f"Key file contains keys which are not a part of the `all_participants` parameter; keys={unwanted_keys}"
                )
            self.service.number_of_agents = len(keys)

        if self.service.number_of_agents > len(keys):
            raise NotValidKeysFile(
                "Number of agents cannot be greater than available keys."
            )

        self._keys = keys

    @staticmethod
    def _get_config_from_json_path(
        override_dict: Dict,
        json_path: Tuple[str, ...],
    ) -> Dict:
        """Returns `setup` param dict"""

        _key_0, *_keys = json_path
        config = override_dict[_key_0]
        for key in _keys:
            config = config[key]
        return config

    @classmethod
    def _try_update_setup_data(
        cls,
        data: List[Tuple[str, Any]],
        override: Dict,
        skill_id: PublicId,
        has_multiple_overrides: bool = False,
    ) -> None:
        """Try update the setup parameters"""

        def _try_update_setup_params(
            setup_param: Optional[Dict],
            setup_data: List[Tuple[str, Any]],
        ) -> None:
            """Update"""
            if setup_param is None:  # pragma: nocover
                return

            for param, value in setup_data:
                setup_param[param] = value

        try:
            if not has_multiple_overrides:
                setup_param = cls._get_config_from_json_path(
                    override_dict=override, json_path=SETUP_PARAM_PATH
                )
                _try_update_setup_params(setup_param=setup_param, setup_data=data)
                return

            for agent_idx in override:
                setup_param = cls._get_config_from_json_path(
                    override_dict=override[agent_idx], json_path=SETUP_PARAM_PATH
                )
                _try_update_setup_params(setup_param=setup_param, setup_data=data)
        except KeyError:
            logging.warning(
                f"Could not update the setup parameter for {skill_id}; "
                f"Configuration does not contain the json path to the setup parameter"
            )

    def _try_update_tendermint_params(
        self,
        override: Dict,
        skill_id: PublicId,
        has_multiple_overrides: bool = False,
    ) -> None:
        """Try update the tendermint parameters"""

        is_kubernetes_deployment = self.deplopyment_type == KUBERNETES_DEPLOYMENT

        def _update_tendermint_params(
            param_args: Dict,
            idx: int,
            is_kubernetes_deployment: bool = False,
        ) -> None:
            """Update tendermint params"""
            if is_kubernetes_deployment:
                param_args[TENDERMINT_URL_PARAM] = TENDERMINT_NODE_LOCAL
                param_args[TENDERMINT_COM_URL_PARAM] = TENDERMINT_COM_LOCAL
            else:
                param_args[TENDERMINT_URL_PARAM] = TENDERMINT_NODE.format(
                    host=self.get_tm_container_name(index=idx)
                )
                param_args[TENDERMINT_COM_URL_PARAM] = TENDERMINT_COM.format(
                    host=self.get_tm_container_name(index=idx)
                )

            if TENDERMINT_P2P_URL_PARAM not in param_args:
                tm_p2p_url = os.environ.get(
                    TENDERMINT_P2P_URL_ENV_VAR.format(idx),
                    TENDERMINT_P2P_URL.format(
                        host=self.get_tm_container_name(index=idx),
                        port=TENDERMINT_P2P_PORT,
                    ),
                )
                param_args[TENDERMINT_P2P_URL_PARAM] = tm_p2p_url

        try:
            if self.service.number_of_agents == 1 and not has_multiple_overrides:
                param_args = self._get_config_from_json_path(
                    override_dict=override, json_path=PARAM_ARGS_PATH
                )
                _update_tendermint_params(
                    param_args=param_args,
                    idx=0,
                    is_kubernetes_deployment=is_kubernetes_deployment,
                )
                return

            if not has_multiple_overrides:
                _base_overrride = deepcopy(override)
                override.clear()
                for i in range(self.get_maximum_participants()):
                    override[i] = deepcopy(_base_overrride)

            for agent_idx in override:
                param_args = self._get_config_from_json_path(
                    override_dict=override[agent_idx], json_path=PARAM_ARGS_PATH
                )
                _update_tendermint_params(
                    param_args=param_args,
                    idx=agent_idx,
                    is_kubernetes_deployment=is_kubernetes_deployment,
                )
        except KeyError:  # pragma: nocover
            logging.warning(
                f"Could not update the tendermint parameter for {skill_id}; "
                f"Configuration does not contain the json path to the setup parameter"
            )

    def try_update_runtime_params(
        self,
        multisig_address: Optional[str] = None,
        agent_instances: Optional[List[str]] = None,
        consensus_threshold: Optional[int] = None,
    ) -> None:
        """Try and update setup parameters."""

        param_overrides: List[Tuple[str, Any]] = []
        if multisig_address is not None:
            param_overrides.append(
                (SAFE_CONTRACT_ADDRESS, multisig_address),
            )

        if agent_instances is not None:
            param_overrides.append(
                (ALL_PARTICIPANTS, agent_instances),
            )

        if consensus_threshold is not None:
            param_overrides.append(
                (CONSENSUS_THRESHOLD, consensus_threshold),
            )

        overrides = copy(self.service.overrides)
        for override in overrides:
            (
                override,
                component_id,
                has_multiple_overrides,
            ) = self.service.process_metadata(
                configuration=override,
            )

            if component_id.component_type.value == SKILL:
                self._try_update_setup_data(
                    data=param_overrides,
                    override=override,
                    skill_id=component_id.public_id,
                    has_multiple_overrides=has_multiple_overrides,
                )
                self._try_update_tendermint_params(
                    override=override,
                    skill_id=component_id.public_id,
                    has_multiple_overrides=has_multiple_overrides,
                )

            override["type"] = component_id.package_type.value
            override["public_id"] = str(component_id.public_id)

        self.service.overrides = overrides

    def get_maximum_participants(self) -> int:
        """Returns the maximum number of participants"""

        if self._all_participants is not None and len(self._all_participants) > 0:
            return len(self._all_participants)

        if self.agent_instances is not None:
            return len(self.agent_instances)

        return max(len(self.keys), self.service.number_of_agents)

    def _update_abci_connection_config(
        self,
        overrides: Dict,
        has_multiple_overrides: bool = False,
    ) -> Dict:
        """Update ABCI connection config"""

        processed_overrides = deepcopy(overrides)
        if self.service.number_of_agents == 1:
            processed_overrides["config"]["host"] = (
                LOCALHOST
                if self.deplopyment_type == KUBERNETES_DEPLOYMENT
                else self.get_abci_container_name(index=0)
            )
            processed_overrides["config"]["port"] = processed_overrides["config"].get(
                "port", DEFAULT_ABCI_PORT
            )
            return processed_overrides

        if not has_multiple_overrides:
            processed_overrides = {}
            for i in range(self.get_maximum_participants()):
                processed_overrides[i] = deepcopy(overrides)

        for idx, override in processed_overrides.items():
            override["config"]["host"] = (
                LOCALHOST
                if self.deplopyment_type == KUBERNETES_DEPLOYMENT
                else self.get_abci_container_name(index=idx)
            )
            override["config"]["port"] = override["config"].get(
                "port", DEFAULT_ABCI_PORT
            )
            processed_overrides[idx] = override

        return processed_overrides

    def try_update_abci_connection_params(
        self,
    ) -> None:
        """Try and update ledger connection parameters."""

        for override in deepcopy(self.service.overrides):
            (
                override,
                component_id,
                has_multiple_overrides,
            ) = self.service.process_metadata(
                configuration=override,
            )

            if (
                component_id.package_type == PackageType.CONNECTION
                and component_id.name == ABCI
            ):
                service_has_connection_overrides = True
                abci_connection_overrides = override
                break
        else:
            service_has_connection_overrides = False
            (
                abci_connection_overrides,
                component_id,
                has_multiple_overrides,
            ) = self.service.process_metadata(
                configuration={
                    "public_id": "valory/abci:0.1.0",
                    "type": PackageType.CONNECTION.value,
                    "config": {
                        "host": LOCALHOST,
                        "port": DEFAULT_ABCI_PORT,
                    },
                }
            )
        processed_overrides = self._update_abci_connection_config(
            overrides=abci_connection_overrides,
            has_multiple_overrides=has_multiple_overrides,
        )

        processed_overrides["public_id"] = str(component_id.public_id)
        processed_overrides["type"] = PackageType.CONNECTION.value

        service_overrides = deepcopy(self.service.overrides)
        if service_has_connection_overrides:
            service_overrides = [
                override
                for override in service_overrides
                if override["public_id"] != str(component_id.public_id)
                and override["type"] != PackageType.CONNECTION.value
            ]

        service_overrides.append(processed_overrides)
        self.service.overrides = service_overrides

    def process_component_overrides(self, agent_n: int) -> Dict:
        """Generates env vars based on model overrides."""
        final_overrides = {}
        for component_configuration_json in self.service.overrides:
            env_var_dict = self.service.process_component_overrides(
                agent_n, component_configuration_json
            )
            final_overrides.update(env_var_dict)
        return final_overrides

    def generate_agents(self) -> List:
        """Generate multiple agent."""
        if self._all_participants is not None and len(self._all_participants) > 0:
            idx_mappings = {
                address: i for i, address in enumerate(self._all_participants)
            }
            agent_override_idx = [
                (i, idx_mappings[kp["address"]]) for i, kp in enumerate(self.keys)
            ]
            return [self.generate_agent(i, idx) for i, idx in agent_override_idx]

        if self.agent_instances is None:
            return [
                self.generate_agent(i) for i in range(self.service.number_of_agents)
            ]

        idx_mappings = {address: i for i, address in enumerate(self.agent_instances)}
        agent_override_idx = [
            (i, idx_mappings[kp["address"]]) for i, kp in enumerate(self.keys)
        ]
        return [self.generate_agent(i, idx) for i, idx in agent_override_idx]

    def generate_common_vars(self, agent_n: int) -> Dict:
        """Retrieve vars common for agent."""
        agent_vars = {
            ENV_VAR_ID: agent_n,
            ENV_VAR_AEA_AGENT: self.service.agent,
            ENV_VAR_LOG_LEVEL: self.log_level,
        }

        if self.private_keys_password is not None:
            agent_vars[ENV_VAR_AEA_PASSWORD] = self.private_keys_password

        return agent_vars

    def generate_agent(
        self, agent_n: int, override_idx: Optional[int] = None
    ) -> Dict[Any, Any]:
        """Generate next agent."""
        agent_vars = self.generate_common_vars(agent_n)
        if len(self.service.overrides) == 0:
            return agent_vars

        if override_idx is None:
            override_idx = agent_n

        agent_vars.update(
            self.process_component_overrides(agent_n=override_idx),
        )
        return agent_vars


class BaseDeploymentGenerator(abc.ABC):  # pylint: disable=too-many-instance-attributes
    """Base Deployment Class."""

    service_builder: ServiceBuilder
    output_name: str
    deployment_type: str
    build_dir: Path
    output: str
    tendermint_job_config: Optional[str]
    dev_mode: bool
    use_tm_testnet_setup: bool

    packages_dir: Path
    open_aea_dir: Path
    open_autonomy_dir: Path

    def __init__(  # pylint: disable=too-many-arguments
        self,
        service_builder: ServiceBuilder,
        build_dir: Path,
        use_tm_testnet_setup: bool = False,
        dev_mode: bool = False,
        packages_dir: Optional[Path] = None,
        open_aea_dir: Optional[Path] = None,
        open_autonomy_dir: Optional[Path] = None,
        image_author: Optional[str] = None,
    ):
        """Initialise with only kwargs."""

        self.service_builder = service_builder
        self.build_dir = build_dir
        self.use_tm_testnet_setup = use_tm_testnet_setup
        self.dev_mode = dev_mode
        self.packages_dir = packages_dir or Path.cwd().absolute() / "packages"
        self.open_aea_dir = open_aea_dir or Path.home().absolute() / "open-aea"
        self.open_autonomy_dir = (
            open_autonomy_dir or Path.home().absolute() / "open-autonomy"
        )

        self.tendermint_job_config: Optional[str] = None
        self.image_author = image_author or DEFAULT_DOCKER_IMAGE_AUTHOR

    @abc.abstractmethod
    def generate(
        self,
        image_version: Optional[str] = None,
        use_hardhat: bool = False,
        use_acn: bool = False,
    ) -> "BaseDeploymentGenerator":
        """Generate the deployment configuration."""

    @abc.abstractmethod
    def generate_config_tendermint(self) -> "BaseDeploymentGenerator":
        """Generate the deployment configuration."""

    @abc.abstractmethod
    def populate_private_keys(
        self,
    ) -> "BaseDeploymentGenerator":
        """Populate the private keys to the deployment."""

    def write_config(self) -> "BaseDeploymentGenerator":
        """Write output to build dir"""

        with open(self.build_dir / self.output_name, "w", encoding="utf8") as f:
            f.write(self.output)
        return self



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for various transaction settlement skill's failure modes."""


import asyncio
import os
import tempfile
import time
from abc import ABC
from pathlib import Path
from threading import Thread
from typing import Any, Callable, Dict, List, Optional, Tuple, cast

from aea.crypto.wallet import Wallet
from aea.decision_maker.base import DecisionMaker
from aea.decision_maker.default import DecisionMakerHandler
from aea.identity.base import Identity
from aea.mail.base import Envelope
from aea.multiplexer import Multiplexer
from aea.protocols.base import Address, Message
from aea.skills.base import Handler
from web3 import HTTPProvider, Web3
from web3.providers import BaseProvider

from packages.valory.skills.abstract_round_abci.base import BaseSynchronizedData
from packages.valory.skills.abstract_round_abci.behaviour_utils import BaseBehaviour
from packages.valory.skills.abstract_round_abci.handlers import SigningHandler
from packages.valory.skills.abstract_round_abci.test_tools.base import (
    FSMBehaviourBaseCase,
)


# pylint: disable=protected-access,too-many-ancestors,unbalanced-tuple-unpacking,too-many-locals,consider-using-with,unspecified-encoding,too-many-arguments,unidiomatic-typecheck

HandlersType = List[Optional[Handler]]
ExpectedContentType = List[
    Optional[
        Dict[
            str,
            Any,
        ]
    ]
]
ExpectedTypesType = List[
    Optional[
        Dict[
            str,
            Any,
        ]
    ]
]


class IntegrationBaseCase(FSMBehaviourBaseCase, ABC):
    """Base test class for integration tests."""

    running_loop: asyncio.AbstractEventLoop
    thread_loop: Thread
    multiplexer: Multiplexer
    decision_maker: DecisionMaker
    agents: Dict[str, Address] = {
        "0xBcd4042DE499D14e55001CcbB24a551F3b954096": "0xf214f2b2cd398c806f84e317254e0f0b801d0643303237d97a22a48e01628897",
        "0x71bE63f3384f5fb98995898A86B02Fb2426c5788": "0x701b615bbdfb9de65240bc28bd21bbc0d996645a3dd57e7b12bc2bdf6f192c82",
        "0xFABB0ac9d68B0B445fB7357272Ff202C5651694a": "0xa267530f49f8280200edf313ee7af6b827f2a8bce2897751d06a843f644967b1",
        "0x1CBd3b2770909D4e10f157cABC84C7264073C9Ec": "0x47c99abed3324a2707c28affff1267e45918ec8c3f20b8aa892e8b065d2942dd",
    }
    current_agent: Address
    ROOT_DIR: Path
    make_ledger_api_connection_callable: Callable

    @classmethod
    def _setup_class(cls, **kwargs: Any) -> None:
        """Setup class."""

    @classmethod
    def setup_class(cls, **kwargs: Any) -> None:
        """Setup."""
        super().setup_class()

        # set up a multiplexer with the required connections
        cls.running_loop = asyncio.new_event_loop()
        cls.thread_loop = Thread(target=cls.running_loop.run_forever)
        cls.thread_loop.start()
        cls.multiplexer = Multiplexer(
            [cls.make_ledger_api_connection_callable()], loop=cls.running_loop
        )
        cls.multiplexer.connect()

        # hardhat configuration
        # setup decision maker
        with tempfile.TemporaryDirectory() as temp_dir:
            fp = os.path.join(temp_dir, "key.txt")
            f = open(fp, "w")
            f.write(cls.agents[next(iter(cls.agents))])
            f.close()
            wallet = Wallet(private_key_paths={"ethereum": str(fp)})
        identity = Identity(
            "test_agent_name",
            addresses=wallet.addresses,
            public_keys=wallet.public_keys,
            default_address_key="ethereum",
        )
        cls._skill._skill_context._agent_context._identity = identity
        cls.current_agent = identity.address

        cls.decision_maker = DecisionMaker(
            decision_maker_handler=DecisionMakerHandler(identity, wallet, {})
        )
        cls._skill._skill_context._agent_context._decision_maker_message_queue = (
            cls.decision_maker.message_in_queue
        )
        cls._skill.skill_context._agent_context._decision_maker_address = (
            "decision_maker"
        )

    @classmethod
    def teardown_class(cls) -> None:
        """Tear down the multiplexer."""
        cls.multiplexer.disconnect()
        cls.running_loop.call_soon_threadsafe(cls.running_loop.stop)
        cls.thread_loop.join()
        super().teardown_class()

    def get_message_from_decision_maker_inbox(self) -> Optional[Message]:
        """Get message from decision maker inbox."""
        if self._skill.skill_context.decision_maker_message_queue.empty():
            return None
        return self._skill.skill_context.decision_maker_message_queue.protected_get(
            self.decision_maker._queue_access_code, block=True
        )

    def process_message_cycle(
        self,
        handler: Optional[Handler] = None,
        expected_content: Optional[Dict] = None,
        expected_types: Optional[Dict] = None,
        mining_interval_secs: float = 0,
    ) -> Optional[Message]:
        """
        Processes one request-response type message cycle.

        Steps:
        1. Calls act on behaviour to generate outgoing message
        2. Checks for message in outbox
        3. Sends message to multiplexer and waits for response.
        4. Passes message to handler
        5. Calls act on behaviour to process incoming message

        :param handler: the handler to handle a potential incoming message
        :param expected_content: the content to be expected
        :param expected_types: the types to be expected
        :param mining_interval_secs: the mining interval used in the tests
        :return: the incoming message
        """
        if expected_types and tuple(expected_types)[0] == "transaction_receipt":
            time.sleep(mining_interval_secs)  # pragma: no cover
        self.behaviour.act_wrapper()
        incoming_message = None

        if type(handler) == SigningHandler:
            self.assert_quantity_in_decision_making_queue(1)
            message = self.get_message_from_decision_maker_inbox()
            assert message is not None, "No message in outbox."  # nosec
            self.decision_maker.handle(message)
            if handler is not None:
                incoming_message = self.decision_maker.message_out_queue.get(block=True)
                assert isinstance(incoming_message, Message)  # nosec
        else:
            self.assert_quantity_in_outbox(1)
            message = self.get_message_from_outbox()
            assert message is not None, "No message in outbox."  # nosec
            self.multiplexer.put(
                Envelope(
                    to=message.to,
                    sender=message.sender,
                    message=message,
                    context=None,
                )
            )
            if handler is not None:
                envelope = self.multiplexer.get(block=True)
                assert envelope is not None, "No envelope"  # nosec
                incoming_message = envelope.message
                assert isinstance(incoming_message, Message)  # nosec

        if handler is not None:
            assert incoming_message is not None  # nosec
            if expected_content is not None:
                assert all(  # nosec
                    [
                        incoming_message._body.get(key, None) == value
                        for key, value in expected_content.items()
                    ]
                ), f"Actual content: {incoming_message._body}, expected: {expected_content}"

            if expected_types is not None:
                assert all(  # nosec
                    [
                        type(incoming_message._body.get(key, None)) == value_type
                        for key, value_type in expected_types.items()
                    ]
                ), "Content type mismatch"
            handler.handle(incoming_message)
            return incoming_message
        return None

    def process_n_messages(
        self,
        ncycles: int,
        synchronized_data: Optional[BaseSynchronizedData] = None,
        behaviour_id: Optional[str] = None,
        handlers: Optional[HandlersType] = None,
        expected_content: Optional[ExpectedContentType] = None,
        expected_types: Optional[ExpectedTypesType] = None,
        fail_send_a2a: bool = False,
        mining_interval_secs: float = 0,
    ) -> Tuple[Optional[Message], ...]:
        """
        Process n message cycles.

        :param behaviour_id: the behaviour to fast forward to
        :param ncycles: the number of message cycles to process
        :param synchronized_data: a synchronized_data
        :param handlers: a list of handlers
        :param expected_content: the expected_content
        :param expected_types: the expected type
        :param fail_send_a2a: flag that indicates whether we want to simulate a failure in the `send_a2a_transaction`
        :param mining_interval_secs: the mining interval used in the tests.

        :return: tuple of incoming messages
        """
        handlers = [None] * ncycles if handlers is None else handlers
        expected_content = (
            [None] * ncycles if expected_content is None else expected_content
        )
        expected_types = [None] * ncycles if expected_types is None else expected_types
        assert (  # nosec
            len(expected_content) == len(expected_types)
            and len(expected_content) == len(handlers)
            and len(expected_content) == ncycles
        ), "Number of cycles, handlers, contents and types does not match"

        if behaviour_id is not None and synchronized_data is not None:
            self.fast_forward_to_behaviour(
                behaviour=self.behaviour,
                behaviour_id=behaviour_id,
                synchronized_data=synchronized_data,
            )
            assert (  # nosec
                cast(BaseBehaviour, self.behaviour.current_behaviour).behaviour_id
                == behaviour_id
            )

        incoming_messages = []
        for i in range(ncycles):
            incoming_message = self.process_message_cycle(
                handlers[i],
                expected_content[i],
                expected_types[i],
                mining_interval_secs,
            )
            incoming_messages.append(incoming_message)

        self.behaviour.act_wrapper()
        if not fail_send_a2a:
            self.mock_a2a_transaction()
        return tuple(incoming_messages)


class HardHatHelperIntegration(IntegrationBaseCase, ABC):  # pragma: no cover
    """Base test class for integration tests with HardHat provider."""

    hardhat_provider: BaseProvider

    @classmethod
    def setup_class(cls, **kwargs: Any) -> None:
        """Setup."""
        super().setup_class()

        # create an API for HardHat
        cls.hardhat_provider = Web3(
            provider=HTTPProvider("http://localhost:8545")
        ).provider



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022-2023 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the links in the documentation."""


import re
import sys
from concurrent.futures import ThreadPoolExecutor
from itertools import chain
from pathlib import Path
from typing import Any, Dict, List, Optional

import requests
import urllib3  # type: ignore
from requests.adapters import HTTPAdapter  # type: ignore
from requests.packages.urllib3.util.retry import (  # type: ignore # pylint: disable=import-error
    Retry,
)


# Disable insecure request warning (expired SSL certificates)
urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


MAX_WORKERS = 10
URL_REGEX = r'(https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s})"]{2,}|www\.[a-zA-Z0-9][a-zA-Z0-9-]+[a-zA-Z0-9]\.[^\s})"]{2,}|https?:\/\/(?:www\.|(?!www))[a-zA-Z0-9]+\.[^\s)"}]{2,}|www\.[a-zA-Z0-9]+\.[^\s)"}]{2,})'
DEFAULT_REQUEST_TIMEOUT = 5  # seconds

# Allow some links to be HTTP because there is no HTTPS alternative
# Remove non-url-allowed characters like ` before adding them here
HTTP_SKIPS = [
    "http://www.fipa.org/repository/ips.php3",
    "http://host.docker.internal:8545",  # internal (ERR_NAME_NOT_RESOLVED)
]

# Special links that are allowed to respond with an error status
# Remove non-url-allowed characters like ` before adding them here
URL_SKIPS = [
    "https://gateway.autonolas.tech/ipfs/<hash>,",  # non link (400)
    "https://github.com/valory-xyz/open-autonomy/trunk/infrastructure",  # svn link (404)
    "http://host.docker.internal:8545",  # internal (ERR_NAME_NOT_RESOLVED)
    "https://github.com/valory-xyz/open-operator",  # Until public
]

# Define here custom timeouts for some edge cases
CUSTOM_TIMEOUTS = {
    "http://www.fipa.org/repository/ips.php3": 30,
}


def read_file(filepath: str) -> str:
    """Loads a file into a string"""
    with open(filepath, "r", encoding="utf-8") as file_:
        file_str = file_.read()
    return file_str


def check_file(
    session: Any,
    md_file: str,
    http_skips: Optional[List[str]] = None,
    url_skips: Optional[List[str]] = None,
) -> Dict:
    """Check for broken or HTTP links in a specific file"""

    http_skips = http_skips or HTTP_SKIPS
    url_skips = url_skips or URL_SKIPS

    text = read_file(md_file)
    m = re.findall(URL_REGEX, text)
    http_links = []
    broken_links = []

    for url in m:
        # Add the closing parenthesis if it is missing, as the REGEX is too strict sometimes
        if "(" in url and ")" not in url:
            url += ")"

        # Remove non allowed chars
        url = url.replace("`", "")

        # Check for HTTP urls
        if not url.startswith("https") and url not in http_skips:
            http_links.append((md_file, url))

        # Check for url skips
        if url in url_skips:
            continue

        # Check for broken links: 200 and 403 codes are admitted
        try:
            # Do not verify requests. Expired SSL certificates would make those links fail
            status_code = session.get(
                url,
                timeout=CUSTOM_TIMEOUTS.get(url, DEFAULT_REQUEST_TIMEOUT),
                verify=False,
            ).status_code
            if status_code not in [200, 403]:
                broken_links.append({"url": url, "status_code": status_code})
        except (
            requests.exceptions.RetryError,
            requests.exceptions.ConnectionError,
        ) as e:
            broken_links.append({"url": url, "status_code": e})

    return {
        "file": str(md_file),
        "http_links": http_links,
        "broken_links": broken_links,
        "n_links": len(m),
    }


def main() -> None:  # pylint: disable=too-many-locals
    """Check for broken or HTTP links"""
    all_md_files = [
        str(p.relative_to("."))
        for p in chain(
            Path("docs").rglob("*.md"),
            Path("packages").rglob("*.md"),
            Path(".").glob("*.md"),
        )
    ]

    broken_links: Dict[str, Dict] = {}
    http_links: Dict[str, List[str]] = {}
    n_links = 0

    # Configure request retries
    retry_strategy = Retry(
        total=3,  # number of retries
        status_forcelist=[404, 429, 500, 502, 503, 504],  # codes to retry on
    )
    # https://stackoverflow.com/questions/18466079/change-the-connection-pool-size-for-pythons-requests-module-when-in-threading
    adapter = HTTPAdapter(
        max_retries=retry_strategy, pool_connections=100, pool_maxsize=100
    )
    session = requests.Session()
    session.mount("https://", adapter)
    session.mount("http://", adapter)

    # Run all file checks in a thread pool
    with ThreadPoolExecutor(max_workers=MAX_WORKERS) as executor:
        futures = []
        for md_file in all_md_files:
            print(f"Checking {str(md_file)}...")
            futures.append(executor.submit(check_file, session, md_file))

        # Awaiting for results is blocking
        print("Awaiting for results...")
        future_results = [future.result() for future in futures]

        # Get errors
        for i in future_results:
            if i["http_links"]:
                http_links[i["file"]] = i["http_links"]
            if i["broken_links"]:
                broken_links[i["file"]] = i["broken_links"]
            if i["n_links"]:
                n_links += i["n_links"]

        print(f"Checked {n_links} links in {len(all_md_files)} files")

        # Check errors
        if broken_links:
            broken_links_str = "\n".join(
                [
                    f"{file_name}: {[entry['url'] + ', status: ' + str(entry['status_code']) for entry in error_data]}"
                    for file_name, error_data in broken_links.items()
                ]
            )
            print(f"Found broken url in the docs:\n{broken_links_str}")

        if http_links:
            http_links_str = "\n".join(
                [
                    f"{file_name}: {[url[1] for url in urls]}"
                    for file_name, urls in http_links.items()
                ]
            )
            print(
                f"Found HTTP urls in the docs:\n{http_links_str}\nTry to use HTTPS equivalent urls or add them to 'http_skips' if not possible"
            )

        if broken_links or http_links:
            sys.exit(1)

        print("OK")
        sys.exit(0)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2023 dummy_author
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the scaffold message definition."""

from enum import Enum
from typing import Any

from aea.configurations.base import PublicId
from aea.exceptions import enforce
from aea.protocols.base import Message
from aea.protocols.scaffold.serialization import MyScaffoldSerializer


class MyScaffoldMessage(Message):
    """The scaffold message class."""

    protocol_id = PublicId.from_str("open_aea/scaffold:0.1.0")
    serializer = MyScaffoldSerializer

    class Performative(Enum):
        """Scaffold Message types."""

        def __str__(self) -> str:
            """Get string representation."""
            return str(self.value)  # pragma: no cover

    def __init__(self, performative: Performative, **kwargs: Any) -> None:
        """
        Initialize.

        :param performative: the type of message.
        :param kwargs: the keyword arguments.
        """
        super().__init__(performative=performative, **kwargs)
        enforce(  # pragma: no cover
            self._is_consistent(), "MyScaffoldMessage initialization inconsistent."
        )

    def _is_consistent(self) -> bool:
        """Check that the data is consistent."""
        try:
            raise NotImplementedError
        except (AssertionError, ValueError):
            return False  # pragma: no cover

        return True  # pragma: no cover
